""" Base for Taggers """
from logging import getLogger


class Tagger:
    """
    Base class for word taggers

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    """

    def __init__(self, config=None, timezone=None, logger=None, **kwargs):
        """
        Parameters
        ----------
        config : Config, default None
            Configuration
        timezone : timezone, default None
            Timezone
        logger : Logger, default None
            Logger
        """
        self.config = config
        self.timezone = timezone
        self.logger = logger or getLogger(__name__)

    def parse(self, text):
        """
        Analyze and parse text

        Parameters
        ----------
        text : str
            Text to analyze

        Returns
        -------
        words : list of minette.WordNode
            Word nodes
        """
        return []



""" Adapter for Clova Extensions Kit """
import traceback

from cek import (
    Clova,
    URL,
    IntentRequest
)

import objson

from .base import Adapter
from ..models import Message


class ClovaAdapter(Adapter):
    """
    Adapter for Clova Extensions Kit

    Attributes
    ----------
    bot : minette.Minette
        Instance of Minette
    application_id : str
        Application ID of Clova Skill
    default_language : str
        Default language of Clova Skill
    clova : Clova
        Clova Extensions Kit API
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    debug : bool
        Debug mode
    """

    def __init__(self, bot=None, *, debug=False,
                 application_id=None, default_language=None, **kwargs):
        """
        Parameters
        ----------
        bot : minette.Minette, default None
            Instance of Minette.
            If None, create new instance of Minette by using `**kwargs`
        application_id : str or None, default None
            Application ID for your Clova Skill
        default_language : str or None, default None
            Default language. ("en" / "ja" / "ko")
            If None, "ja" is set to Clova Extensions Kit API object
        debug : bool, default False
            Debug mode
        """
        super().__init__(bot=bot, threads=0, debug=debug, **kwargs)
        self.application_id = application_id or \
            self.config.get(section="clova_cek", key="application_id")
        self.default_language = default_language or \
            self.config.get(section="clova_cek", key="default_language") or "ja"
        self.clova = Clova(application_id=self.application_id,
                           default_language=self.default_language,
                           debug_mode=debug)

        # handler for all types of request
        @self.clova.handle.default
        def default(clova_request):
            return clova_request

    def handle_http_request(self, request_data, request_headers):
        """
        Interface to chat with Clova Skill

        Parameters
        ----------
        request_data : bytes
            Request data from Clova as bytes
        request_headers : dict
            Request headers from Clova as dict

        Returns
        -------
        response : Response
            Response from chatbot. Send back `json` attribute to Clova API
        """
        clova_request = self.clova.route(request_data, request_headers)
        return self.handle_event(clova_request)

    def handle_event(self, clova_request):
        # execute bot
        channel_messages, _ = super().handle_event(clova_request)

        # print response for debug
        for msg in channel_messages:
            if self.debug:
                self.logger.info(msg)
            else:
                self.logger.info("Minette> {}".format(msg["speech_value"]))

        # build response message
        speech_values = [msg["speech_value"] for msg in channel_messages]
        end_session = channel_messages[-1]["end_session"]
        reprompt = channel_messages[-1]["reprompt"]
        if len(speech_values) == 1:
            return objson.dumps(self.clova.response(
                speech_values[0], end_session=end_session, reprompt=reprompt))
        else:
            return objson.dumps(self.clova.response(
                speech_values, end_session=end_session, reprompt=reprompt))

    @staticmethod
    def _to_minette_message(clova_request):
        """
        Convert ClovaRequest object to Minette Message object

        Parameters
        ----------
        clova_request : cek.Request
            Request from clova

        Returns
        -------
        message : minette.Message
            Request converted into Message object
        """
        msg = Message(
            type=clova_request.type,
            channel="LINE",
            channel_detail="Clova",
            channel_user_id=clova_request.session.user.id if clova_request.session._session else "",
            channel_message=clova_request
        )

        # Set intent and entities when IntentRequest
        if isinstance(clova_request, IntentRequest):
            msg.intent = clova_request.name
            # if clova_request.slots: <- Error occures when no slot values
            if clova_request._request["intent"]["slots"]:
                msg.entities = clova_request.slots
        return msg

    @staticmethod
    def _to_channel_message(message):
        """
        Convert Minette Message object to LINE SendMessage object

        Parameters
        ----------
        response : Message
            Response message object

        Returns
        -------
        response : SendMessage
            SendMessage object for LINE Messaging API
        """
        return {
            "speech_value": URL(message.text) if message.type == "url" else message.text,
            "end_session": message.entities.get("end_session", True),
            "reprompt": message.entities.get("reprompt", None)
        }



""" Base class for DialogService for processing each dialogs """
import traceback
from logging import Logger, getLogger

from ..models import (
    Message,
    Response,
    Context,
    PerformanceInfo
)


class DialogService:
    """
    Base class for DialogService

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : timezone
        Timezone
    logger : logging.Logger
        Logger
    """

    @classmethod
    def topic_name(cls):
        """
        Topic name of this dialog service

        Returns
        -------
        topic_name : str
            Topic name of this dialog service
        """
        cls_name = cls.__name__.lower()
        if cls_name.endswith("dialogservice"):
            cls_name = cls_name[:-13]
        elif cls_name.endswith("dialog"):
            cls_name = cls_name[:-6]
        return cls_name

    def __init__(self, config=None, timezone=None, logger=None):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        """
        self.config = config
        self.timezone = timezone
        self.logger = logger or getLogger(__name__)

    def execute(self, request, context, connection, performance):
        """
        Main logic of DialogService

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection
        performance : minette.PerformanceInfo
            Performance information

        Returns
        -------
        response : minette.Response
            Response from chatbot
        """
        try:
            # extract entities
            for k, v in self.extract_entities(
                    request, context, connection).items():
                if not request.entities.get(k, ""):
                    request.entities[k] = v
            performance.append("dialog_service.extract_entities")

            # initialize context data
            if context.topic.is_new:
                context.data = self.get_slots(request, context, connection)
            performance.append("dialog_service.get_slots")

            # process request
            self.process_request(request, context, connection)
            performance.append("dialog_service.process_request")

            # compose response
            response_messages = \
                self.compose_response(request, context, connection)
            if not response_messages:
                self.logger.info("No response")
                response_messages = []
            elif not isinstance(response_messages, list):
                response_messages = [response_messages]
            response = Response()
            for rm in response_messages:
                if isinstance(rm, Message):
                    response.messages.append(rm)
                elif isinstance(rm, str):
                    response.messages.append(request.to_reply(text=rm))
            performance.append("dialog_service.compose_response")

        except Exception as ex:
            self.logger.error(
                "Error occured in dialog_service: "
                + str(ex) + "\n" + traceback.format_exc())
            response = Response(messages=[
                self.handle_exception(request, context, ex, connection)])

        return response

    def extract_entities(self, request, context, connection):
        """
        Extract entities from request message

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection

        Returns
        -------
        entities : dict
            Entities extracted from request message
        """
        return {}

    def get_slots(self, request, context, connection):
        """
        Get initial context.data at the start of this dialog

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection

        Returns
        -------
        slots : dict
            Initial context.data
        """
        return {}

    def process_request(self, request, context, connection):
        """
        Process your chatbot's functions/skills and setup context data

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection
        """
        pass

    def compose_response(self, request, context, connection):
        """
        Compose response messages using context data

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection

        Returns
        -------
        response : minette.Response
            Response from chatbot
        """
        return ""

    def handle_exception(self, request, context, exception, connection):
        """
        Handle exception and return error response message

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        exception : Exception
            Exception
        connection : Connection
            Connection

        Returns
        -------
        response : minette.Response
            Error response from chatbot
        """
        context.set_error(exception)
        context.topic.keep_on = False
        return request.to_reply(text="?")


class EchoDialogService(DialogService):
    """
    Simple echo dialog service for tutorial

    """
    def compose_response(self, request, context, connection=None):
        return request.to_reply(text="You said: {}".format(request.text))


class ErrorDialogService(DialogService):
    """
    Dialog service for error in chatting

    """
    def compose_response(self, request, context, connection=None):
        return request.to_reply(text="?")



""" Base class for DialogRouter that routes proper dialog for the intent """
from abc import ABC, abstractmethod
import traceback
from logging import Logger, getLogger

from ..models import Message, Priority
from .service import DialogService, ErrorDialogService


class DialogRouter:
    """
    Base class for DialogRouter

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : timezone
        Timezone
    logger : logging.Logger
        Logger
    default_dialog_service : DialogService
        Dialog service used when intent is not clear
    intent_resolver : dict
        Resolver for intent to dialog
    topic_resolver : dict
        Resolver for topic to dialog for successive chatting
    """

    def __init__(self, config=None, timezone=None, logger=None,
                 default_dialog_service=None, intent_resolver=None, **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        default_dialog_service : minette.DialogService or type, default None
            Dialog service used when intent is not clear.
        """
        self.config = config
        self.timezone = timezone
        self.logger = logger or getLogger(__name__)
        self.default_dialog_service = default_dialog_service or DialogService
        # set up intent_resolver
        self.intent_resolver = intent_resolver or {}
        self.register_intents()
        # set up topic_resolver
        self.topic_resolver = {
            v.topic_name(): v for v in self.intent_resolver.values() if v}
        self.topic_resolver[self.default_dialog_service.topic_name()] = \
            self.default_dialog_service

    def register_intents(self):
        """
        Register intents and the dialog services to process the intents

        >>> self.intent_resolver = {
            "PizzaOrderIntent": PizzaDialogService,
            "ChangeAddressIntent": ChangeAddressDialogService,
        }
        """
        pass

    def execute(self, request, context, connection, performance):
        """
        Main logic of DialogRouter

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection
        performance : minette.PerformanceInfo
            Performance information

        Returns
        -------
        dialog_service : minette.DialogService
            DialogService to process request message
        """
        try:
            # extract intent and entities
            extracted = self.extract_intent(
                request=request, context=context, connection=connection)
            if isinstance(extracted, tuple):
                request.intent = extracted[0]
                request.entities = extracted[1]
                if len(extracted) > 2:
                    request.intent_priority = extracted[2]
            elif isinstance(extracted, str):
                request.intent = extracted
            performance.append("dialog_router.extract_intent")
            # preprocess before route
            self.before_route(request, context, connection)
            performance.append("dialog_router.before_route")
            # route dialog
            dialog_service = self.route(request, context, connection)
            if issubclass(dialog_service, DialogService):
                dialog_service = dialog_service(
                    config=self.config, timezone=self.timezone,
                    logger=self.logger
                )
            performance.append("dialog_router.route")
        except Exception as ex:
            self.logger.error(
                "Error occured in dialog_router: "
                + str(ex) + "\n" + traceback.format_exc())
            dialog_service = \
                self.handle_exception(request, context, ex, connection)

        return dialog_service

    def extract_intent(self, request, context, connection):
        """
        Extract intent and entities from request message

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection

        Returns
        -------
        response : tuple of (str, dict)
            Intent and entities
        """
        return request.intent, request.entities

    def before_route(self, request, context, connection):
        """
        Preprocessing for all requests before routing

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection
        """
        pass

    def route(self, request, context, connection):
        """
        Return proper DialogService for intent or topic

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        connection : Connection
            Connection


        Returns
        -------
        dialog_service : minette.DialogService
            Dialog service proper for intent or topic
        """
        # update
        if request.intent in self.intent_resolver and (
                request.intent_priority > context.topic.priority or
                not context.topic.name):
            dialog_service = self.intent_resolver[request.intent]
            # update topic if request is not adhoc
            if dialog_service and not request.is_adhoc:
                context.topic.name = dialog_service.topic_name()
                context.topic.status = ""
                if request.intent_priority >= Priority.Highest:
                    # set slightly lower priority to enable to update Highest priority intent
                    context.topic.priority = Priority.Highest - 1
                else:
                    context.topic.priority = request.intent_priority
                context.topic.is_new = True
            # do not update topic when request is adhoc or ds is None
            else:
                dialog_service = dialog_service or DialogService
                if context.topic.name:
                    context.topic.keep_on = True

        # continue
        elif context.topic.name:
            dialog_service = self.topic_resolver[context.topic.name]

        # default (intent not extracted or unknown)
        else:
            dialog_service = self.default_dialog_service
            context.topic.name = dialog_service.topic_name()
            context.topic.status = ""
            context.topic.is_new = True
        return dialog_service

    def handle_exception(self, request, context, exception, connection):
        """
        Handle exception and return ErrorDialogService

        Parameters
        ----------
        request : minette.Message
            Request message
        context : minette.Context
            Context
        exception : Exception
            Exception
        connection : Connection
            Connection

        Returns
        -------
        response : minette.ErrorDialogService
            Dialog service for error occured in chatting
        """
        context.set_error(exception)
        return ErrorDialogService(
            config=self.config, timezone=self.timezone, logger=self.logger)



""" Base class for ContextStore """
from abc import ABC, abstractmethod
import traceback
from logging import getLogger
from datetime import datetime
from pytz import timezone as tz

import objson

from ..models import Context, Topic


class ContextStore(ABC):
    """
    Base class for ContextStore to enable successive conversation

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    table_name : str
        Database table name for read/write context data
    sqls : dict
        SQLs used in ContextStore
    timeout : int
        Context timeout (Seconds)
    """

    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="context", *, timeout=300, **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        table_name : str, default "context"
            Database table name for read/write context data
        timeout : int, default 300
            Context timeout (Seconds)
        """
        self.config = config
        self.timezone = timezone or (
            tz(config.get("timezone", default="UTC")) if config else tz("UTC"))
        self.logger = logger if logger else getLogger(__name__)
        self.table_name = table_name
        self.timeout = timeout
        self.sqls = self.get_sqls()

    @abstractmethod
    def get_sqls(self):
        pass

    def prepare_table(self, connection, prepare_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default tuple()
            Query parameters for checking table

        Returns
        -------
        created : bool
            Return True when created new table
        """
        cursor = connection.cursor()
        cursor.execute(self.sqls["prepare_check"], prepare_params or tuple())
        if not cursor.fetchone():
            cursor.execute(self.sqls["prepare_create"])
            connection.commit()
            return True
        else:
            return False

    def get(self, channel, channel_user_id, connection):
        """
        Get context by channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        context : minette.Context
            Context for channel and channel_user_id
        """
        context = Context(channel, channel_user_id)
        context.timestamp = datetime.now(self.timezone)
        if not channel_user_id:
            return context
        try:
            cursor = connection.cursor()
            cursor.execute(
                self.sqls["get_context"], (channel, channel_user_id))
            row = cursor.fetchone()
            if row is not None:
                # convert to dict
                if isinstance(row, dict):
                    record = row
                else:
                    record = dict(
                        zip([column[0] for column in cursor.description], row))
                # convert type
                record["topic_previous"] = \
                    objson.loads(record["topic_previous"])
                record["data"] = objson.loads(record["data"])
                # check context timeout
                if record["timestamp"].tzinfo:
                    last_access = record["timestamp"].astimezone(self.timezone)
                else:
                    last_access = self.timezone.localize(record["timestamp"])
                gap = datetime.now(self.timezone) - last_access
                if gap.total_seconds() <= self.timeout:
                    # restore context if not timeout
                    context.topic.name = record["topic_name"]
                    context.topic.status = record["topic_status"]
                    context.topic.priority = record["topic_priority"]
                    context.topic.previous = Topic.from_dict(
                        record["topic_previous"]) \
                        if record["topic_previous"] else None
                    context.data = record["data"] if record["data"] else {}
                    context.is_new = False
        except Exception as ex:
            self.logger.error(
                "Error occured in restoring context from database: "
                + str(ex) + "\n" + traceback.format_exc())
        return context

    def save(self, context, connection):
        """
        Save context

        Parameters
        ----------
        context : minette.Context
            Context to save
        connection : Connection
            Connection
        """
        if not context.channel_user_id:
            return
        # serialize some elements
        context_dict = context.to_dict()
        serialized_previous_topic = \
            objson.dumps(context_dict["topic"]["previous"])
        serialized_data = objson.dumps(context_dict["data"])
        # save
        cursor = connection.cursor()
        cursor.execute(self.sqls["save_context"], (
            context.channel, context.channel_user_id, context.timestamp,
            context.topic.name, context.topic.status,
            serialized_previous_topic, context.topic.priority,
            serialized_data))
        connection.commit()



class Priority:
    """
    Priority of topic

    Attributes
    ----------
    Highest : int
        Highest (100)
    High : int
        High (75)
    Normal : int
        Normal (50)
    Low : int
        Low (25)
    Ignore : int
        Ignore (0)
    """
    Highest = 100
    High = 75
    Normal = 50
    Low = 25
    Ignore = 0



import traceback
from copy import deepcopy
from objson import Serializable

from .priority import Priority
from .topic import Topic


class Context(Serializable):
    """
    Context

    Attributes
    ----------
    channel : str
        Channel
    channel_user_id : str
        Channel user ID
    timestamp : datetime
        Timestamp
    is_new : bool
        True if context is created at this turn
    topic : Topic
        Current topic
    data : dict
        Data slots
    error : dict
        Error info
    """
    def __init__(self, channel=None, channel_user_id=None):
        """
        Parameters
        ----------
        channel : str, default None
            Channel
        channel_user_id : str, default None
            Channel user ID
        """
        self.channel = channel
        self.channel_user_id = \
            channel_user_id if isinstance(channel_user_id, str) else ""
        self.timestamp = None
        self.is_new = True
        self.topic = Topic()
        self.data = {}
        self.error = {}

    def reset(self, keep_data=False):
        """
        Backup to previous topic and remove data

        Parameters
        ----------
        keep_data : bool, default False
            Keep context data to next turn
        """
        # backup previous topic
        self.topic.previous = None
        self.topic.previous = deepcopy(self.topic)
        # remove data if topic not keep_on
        if not self.topic.keep_on:
            self.topic.name = ""
            self.topic.status = ""
            self.topic.priority = Priority.Normal
            self.data = self.data if keep_data else {}
            self.error = {}

    def set_error(self, ex, info=None):
        """
        Set error info

        Parameters
        ----------
        ex : Exception
            Exception
        info : dict, default None
            More information for debugging
        """
        self.error = {
            "exception": str(ex),
            "traceback": traceback.format_exc(),
            "info": info if info else {}}

    @classmethod
    def _types(cls):
        return {
            "topic": Topic
        }



""" Tagger using mecab-service """
import traceback
import requests

from ..models import WordNode
from .base import Tagger


class MeCabServiceNode(WordNode):
    """
    Parsed word node by MeCabServiceTagger

    Attributes
    ----------
    surface : str
        Surface of word
    part : str
        Part of the word
    part_detail1 : str
        Detail1 of part
    part_detail2 : str
        Detail2 of part
    part_detail3 : str
        Detail3 of part
    stem_type : str
        Stem type
    stem_form : str
        Stem form
    word : str
        Word itself
    kana : str
        Japanese kana of the word
    pronunciation : str
        Pronunciation of the word
    """

    @classmethod
    def create(cls, surface, features):
        """
        Create instance of MeCabServiceNode

        Parameters
        ----------
        surface : str
            Surface of the word
        features : dict
            Features analyzed by MeCabService
        """
        return cls(
            surface=surface,
            part=features["part"],
            part_detail1=features["part_detail1"],
            part_detail2=features["part_detail2"],
            part_detail3=features["part_detail3"],
            stem_type=features["stem_type"],
            stem_form=features["stem_form"],
            word=features["word"],
            kana=features["kana"],
            pronunciation=features["pronunciation"]
        )


class MeCabServiceTagger(Tagger):
    """
    Tagger using mecab-service

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    api_url : str
        URL for MeCabService API
    """

    def __init__(self, config=None, timezone=None, logger=None, *,
                 api_url=None, **kwargs):
        """
        Parameters
        ----------
        config : Config, default None
            Configuration
        timezone : timezone, default None
            Timezone
        logger : Logger, default None
            Logger
        api_url : str, default None
            URL for MeCabService API.
            If None trial URL is used.
        """
        super().__init__(config=config, timezone=timezone, logger=logger)
        if not api_url:
            self.api_url = "https://api.uezo.net/mecab/parse"
            self.logger.warning(
                "Do not use default API URL for the production environment. "
                "This is for trial use only. "
                "Install MeCab and use MeCabTagger instead.")
        else:
            self.api_url = api_url

    def parse(self, text):
        """
        Parse and annotate using MeCab Service

        Parameters
        ----------
        text : str
            Text to analyze

        Returns
        -------
        words : list of minette.MeCabServiceNode
            MeCabService nodes
        """
        ret = []
        if not text:
            return ret
        try:
            parsed_json = requests.post(
                self.api_url, headers={"content-type": "application/json"},
                json={"text": text}, timeout=10).json()
            ret = [MeCabServiceNode.create(
                n["surface"], n["features"]) for n in parsed_json["nodes"]]
        except Exception as ex:
            self.logger.error(
                "MeCab Service parsing error: "
                + str(ex) + "\n" + traceback.format_exc())
        return ret



""" Utilities functions for minette """
from datetime import datetime
import calendar
import objson


def date_to_str(dt, with_timezone=False):
    """
    Convert datetime to str

    Parameters
    ----------
    dt : datetime
        datetime to convert
    with_timezone : bool, default False
        Include timezone or not

    Returns
    -------
    datetime_str : str
        Datetime string
    """
    return objson.date_to_str(dt, with_timezone)


def str_to_date(dtstr):
    """
    Convert str to datetime

    Parameters
    ----------
    dtstr : str
        str to convert

    Returns
    -------
    datetime : datetime
        datetime
    """
    return objson.str_to_date(dtstr)


def date_to_unixtime(dt):
    """
    Convert datetime to unixtime

    Parameters
    ----------
    dt : datetime
        datetime to convert

    Returns
    -------
    unixtime : int
        Unixtime
    """
    return calendar.timegm(dt.utctimetuple())


def unixtime_to_date(unixtime, tz=None):
    """
    Convert unixtime to datetime

    Parameters
    ----------
    unixtime : int
        unixtime to convert
    tz : timezone
        timezone to set

    Returns
    -------
    datetime : datetime
        datetime
    """
    return datetime.fromtimestamp(unixtime, tz=tz)


def encode_json(obj, **kwargs):
    return objson.dumps(obj, **kwargs)


def decode_json(json_string, **kwargs):
    return objson.loads(json_string, **kwargs)



from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet

from .sqlitestores import (
    SQLiteConnectionProvider,
    SQLiteContextStore,
    SQLiteUserStore,
    SQLiteMessageLogStore,
    SQLiteStores
)



import traceback
from datetime import datetime, timezone, MAXYEAR
from time import time

from azure.cosmosdb.table.tableservice import TableService
from azure.common import AzureMissingResourceHttpError

from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet

from ..models import (
    Context,
    Topic,
    User
)

import objson


class AzureTableConnection(TableService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class AzureTableConnectionProvider(ConnectionProvider):
    """
    Connection provider for Azure Table Storage

    Attributes
    ----------
    connection_str : str
        Connection string
    """
    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        return AzureTableConnection(connection_string=self.connection_str)


class AzureTableContextStore(ContextStore):
    """
    Context store for Azure SQL Database to enable successive conversation

    """
    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, query_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default None
            Query parameters for checking table
        """
        if not connection.exists(self.table_name):
            connection.create_table(self.table_name)
            return True
        else:
            return False

    def get(self, channel, channel_user_id, connection):
        """
        Get context for channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        context : Context
            Context for channel and channel_user_id
        """
        context = Context(channel, channel_user_id)
        context.timestamp = datetime.now(self.timezone)
        if not channel_user_id:
            return context
        try:
            row = connection.get_entity(self.table_name, channel, channel_user_id)
            if row is not None:
                # convert type
                row["topic_previous"] = objson.loads(row["topic_previous"])
                row["data"] = objson.loads(row["data"])
                # check context timeout. AzureTableStorage always returns timestamp with UTC timezone
                last_access = row["local_timestamp"].astimezone(self.timezone)
                if (datetime.now(self.timezone) - last_access).total_seconds() <= self.timeout:
                    # restore context if not timeout
                    context.topic.name = row["topic_name"]
                    context.topic.status = row["topic_status"]
                    context.topic.priority = row["topic_priority"]
                    context.topic.previous = Topic.from_dict(row["topic_previous"]) if row["topic_previous"] else None
                    context.data = row["data"] if row["data"] else {}
                    context.is_new = False
        except Exception as ex:
            self.logger.error("Error occured in restoring context from database: " + str(ex) + "\n" + traceback.format_exc())
        return context

    def save(self, context, connection):
        """
        Save context

        Parameters
        ----------
        context : Context
            Context to save
        connection : Connection
            Connection
        """
        if not context.channel_user_id:
            return
        # serialize some elements
        context_dict = context.to_dict()
        serialized_previous_topic = objson.dumps(context_dict["topic"]["previous"])
        serialized_data = objson.dumps(context_dict["data"])
        # save
        entity = {
            "PartitionKey": context.channel,
            "RowKey": context.channel_user_id,
            "local_timestamp": context.timestamp,
            "topic_name": context.topic.name,
            "topic_status": context.topic.status,
            "topic_previous": serialized_previous_topic,
            "topic_priority": context.topic.priority,
            "data": serialized_data
        }
        connection.insert_or_replace_entity(self.table_name, entity)


class AzureTableUserStore(UserStore):
    """
    User store for Azure SQL Database

    """
    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, query_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default None
            Query parameters for checking table
        """
        if not connection.exists(self.table_name):
            connection.create_table(self.table_name)
            return True
        else:
            return False

    def get(self, channel, channel_user_id, connection):
        """
        Get user from repository by channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        user : User
            User
        """
        user = User(channel=channel, channel_user_id=channel_user_id)
        if not channel_user_id:
            return user
        try:
            row = connection.get_entity(self.table_name, channel, channel_user_id)
            # convert type
            row["data"] = objson.loads(row["data"])
            # restore user
            user.id = row["user_id"]
            user.name = row["name"]
            user.nickname = row["nickname"]
            user.profile_image_url = row["profile_image_url"]
            user.data = row["data"] if row["data"] else {}

        except AzureMissingResourceHttpError as amrherr:
            # add new user if user is not found
            if "ResourceNotFound" in amrherr.args[0]:
                entity = {
                    "PartitionKey": channel,
                    "RowKey": channel_user_id,
                    "local_timestamp": datetime.now(self.timezone),
                    "user_id": user.id,
                    "name": user.name,
                    "nickname": user.nickname,
                    "profile_image_url": user.profile_image_url,
                    "data": objson.dumps({})
                }
                connection.insert_entity(self.table_name, entity)
            else:
                self.logger.error("Resouce is missing on Azure Table: " + str(amrherr) + "\n" + traceback.format_exc())

        except Exception as ex:
            print(ex)
            self.logger.error("Error occured in restoring user from Azure Table: " + str(ex) + "\n" + traceback.format_exc())
        return user

    def save(self, user, connection):
        """
        Save user

        Parameters
        ----------
        user : User
            User to save
        connection : Connection
            Connection
        """
        user_dict = user.to_dict()
        serialized_data = objson.dumps(user_dict["data"])
        entity = {
            "PartitionKey": user.channel,
            "RowKey": user.channel_user_id,
            "local_timestamp": datetime.now(self.timezone),
            "user_id": user.id,
            "name": user.name,
            "nickname": user.nickname,
            "profile_image_url": user.profile_image_url,
            "data": serialized_data
        }
        connection.insert_or_replace_entity(self.table_name, entity)


class AzureTableMessageLogStore(MessageLogStore):
    """
    Message log store for Azure SQL Database

    """
    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, query_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default None
            Query parameters for checking table
        """
        if not connection.exists(self.table_name):
            connection.create_table(self.table_name)
            return True
        else:
            return False

    def save(self, request, response, context, connection):
        """
        Write message log

        Parameters
        ----------
        request : Message
            Request message
        response : Response
            Response from chatbot
        context : Context
            Context
        total_ms : int
            Response time (milliseconds)
        connection : Connection
            Connection
        """
        epoch_max = datetime(MAXYEAR, 12, 31, 23, 59, 59, 999999, timezone.utc).timestamp()
        entity = self._flatten(request, response, context)
        entity["PartitionKey"] = request.channel_user_id
        entity["RowKey"] = str(epoch_max - time())
        entity["request_json"] = request.to_json()
        entity["response_json"] = response.to_json()
        entity["context_json"] = context.to_json()
        connection.insert_entity(self.table_name, entity)
        return entity["RowKey"]


class AzureTableStores(StoreSet):
    connection_provider = AzureTableConnectionProvider
    context_store = AzureTableContextStore
    user_store = AzureTableUserStore
    messagelog_store = AzureTableMessageLogStore



import traceback
from datetime import datetime
from copy import deepcopy
import objson

from sqlalchemy import (
    create_engine,
    Column,
    Integer,
    String,
    DateTime,
    Boolean,
    TEXT
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet

from ..models import (
    Context,
    Topic,
    User
)

# Base of models
Base = declarative_base()


# Models

class SQLAlchemyContext(Context, Base):
    __tablename__ = "context"
    channel = Column("channel", String(length=20), primary_key=True)
    channel_user_id = Column("channel_user_id", String(length=100), primary_key=True)
    timestamp = Column("timestamp", DateTime)
    topic_name = Column("topic_name", String(length=100))
    topic_status = Column("topic_status", String(length=100))
    topic_previous = Column("topic_previous", String(length=4000))
    topic_priority = Column("topic_priority", Integer)
    data = Column("data", TEXT)


class SQLAlchemyUser(User, Base):
    __tablename__ = "user"
    channel = Column("channel", String(length=20), primary_key=True)
    channel_user_id = Column("channel_user_id", String(length=100), primary_key=True)
    id = Column("user_id", String(length=100))
    timestamp = Column("timestamp", DateTime)
    name = Column("name", String(length=100))
    nickname = Column("nickname", String(length=100))
    profile_image_url = Column("profile_image_url", String(length=500))
    data = Column("data", TEXT)


class SQLAlchemyMessageLog(objson.Serializable, Base):
    __tablename__ = "messagelog"
    id = Column("id", Integer, primary_key=True, autoincrement=True)
    channel = Column("channel", String(length=20))
    channel_detail = Column("channel_detail", String(length=100))
    channel_user_id = Column("channel_user_id", String(length=100))
    request_timestamp = Column("request_timestamp", DateTime)
    request_id = Column("request_id", String(length=100))
    request_type = Column("request_type", String(length=100))
    request_text = Column("request_text", String(length=4000))
    request_payloads = Column("request_payloads", TEXT)
    request_intent = Column("request_intent", String(length=100))
    request_is_adhoc = Column("request_is_adhoc", Boolean)
    response_type = Column("response_type", String(length=100))
    response_text = Column("response_text", String(length=4000))
    response_payloads = Column("response_payloads", TEXT)
    response_milliseconds = Column("response_milliseconds", Integer)
    context_is_new = Column("context_is_new", Boolean)
    context_topic_name = Column("context_topic_name", String(length=100))
    context_topic_status = Column("context_topic_status", String(length=100))
    context_topic_is_new = Column("context_topic_is_new", Boolean)
    context_topic_keep_on = Column("context_topic_keep_on", Boolean)
    context_topic_priority = Column("context_topic_priority", Integer)
    context_error = Column("context_error", TEXT)
    request_json = Column("request_json", TEXT)
    response_json = Column("response_json", TEXT)
    context_json = Column("context_json", TEXT)


# DataStores

class SQLAlchemyConnection(Session):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class SQLAlchemyConnectionProvider(ConnectionProvider):
    def __init__(self, connection_str, db_engine=None, db_echo=True, **kwargs):
        """
        Parameters
        ----------
        connection_str : str
            Connection string
        """
        super().__init__(connection_str)
        self.engine = db_engine or create_engine(self.connection_str, echo=db_echo)

    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        return SQLAlchemyConnection(
            autocommit=False,
            autoflush=True,
            bind=self.engine)


class SQLAlchemyContextStore(ContextStore):
    """
    Context store for SQLAlchemy to enable successive conversation

    """
    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="context", *, timeout=300, **kwargs):
        super().__init__(config, timezone, logger, table_name,
                 timeout=timeout, **kwargs)
        SQLAlchemyContext.__table__.name = self.table_name

    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, prepare_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default tuple()
            Query parameters for checking table

        Returns
        -------
        created : bool
            Return True when created new table
        """
        Base.metadata.create_all(bind=connection.bind, tables=[SQLAlchemyContext.__table__])
        # Always return `False` because the result can't be handled
        return False

    def get(self, channel, channel_user_id, connection):
        """
        Get context by channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        context : minette.Context
            Context for channel and channel_user_id
        """

        # Create new instance and map stored values to assure that constructor is always called

        context = SQLAlchemyContext(channel, channel_user_id)
        context.timestamp = datetime.now(self.timezone)

        if not channel_user_id:
            return context

        try:
            stored_context = connection.query(SQLAlchemyContext).filter(SQLAlchemyContext.channel==channel, SQLAlchemyContext.channel_user_id==channel_user_id).first()
            if stored_context is not None:
                # check context timeout
                if stored_context.timestamp.tzinfo:
                    last_access = stored_context.timestamp.astimezone(self.timezone)
                else:
                    last_access = self.timezone.localize(stored_context.timestamp)

                gap = datetime.now(self.timezone) - last_access
                if gap.total_seconds() <= self.timeout:
                    # restore context if not timeout
                    context.topic.name = stored_context.topic_name
                    context.topic.status = stored_context.topic_status
                    context.topic.priority = stored_context.topic_priority
                    context.topic.previous = Topic.from_json(
                        stored_context.topic_previous) \
                        if stored_context.topic_previous else None
                    context.data = objson.loads(stored_context.data) \
                        if stored_context.data else {}
                    context.is_new = False

        except Exception as ex:
            self.logger.error(
                "Error occured in restoring context from database: "
                + str(ex) + "\n" + traceback.format_exc())

        return context

    def save(self, context, connection):
        """
        Save context

        Parameters
        ----------
        context : SQLAlchemyContext
            Context to save
        connection : Session
            Connection
        """
        if not context.channel_user_id:
            return

        # copy and serialize values to store
        context_to_store = deepcopy(context)
        context_to_store.topic_name = context_to_store.topic.name
        context_to_store.topic_status = context_to_store.topic.status
        context_to_store.topic_previous = objson.dumps(context_to_store.topic.previous)
        context_to_store.topic_priority = context_to_store.topic.priority
        context_to_store.data = objson.dumps(context_to_store.data)

        # save
        connection.merge(instance=context_to_store)
        connection.commit()


class SQLAlchemyUserStore(UserStore):
    """
    User store for SQLAlchemy

    """
    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="user", **kwargs):
        super().__init__(config, timezone, logger, table_name, **kwargs)
        SQLAlchemyUser.__table__.name = self.table_name

    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, query_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default None
            Query parameters for checking table

        Returns
        -------
        created : bool
            Return True when created new table
        """
        Base.metadata.create_all(bind=connection.bind, tables=[SQLAlchemyUser.__table__])
        # Always return `False` because the result can't be handled
        return False

    def get(self, channel, channel_user_id, connection):
        """
        Get user from repository by channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        user : User
            User
        """

        # Create new instance and map stored values to assure that constructor is always called

        user = SQLAlchemyUser(channel=channel, channel_user_id=channel_user_id)

        if not channel_user_id:
            return user

        try:
            stored_user = connection.query(SQLAlchemyUser).filter(SQLAlchemyUser.channel==channel, SQLAlchemyUser.channel_user_id==channel_user_id).first()

            if stored_user is not None:
                user.id = stored_user.id
                user.name = stored_user.name
                user.nickname = stored_user.nickname
                user.profile_image_url = stored_user.profile_image_url
                user.data = objson.loads(stored_user.data) if stored_user.data else {}

            else:
                self.save(user, connection)

        except Exception as ex:
            self.logger.error(
                "Error occured in restoring user from database: "
                + str(ex) + "\n" + traceback.format_exc())

        return user

    def save(self, user, connection):
        """
        Save user

        Parameters
        ----------
        user : User
            User to save
        connection : Connection
            Connection
        """
        # copy and serialize values to store
        user_to_store = deepcopy(user)
        user_to_store.timestamp = datetime.now(self.timezone)
        user_to_store.data = objson.dumps(user_to_store.data)

        # save
        connection.merge(instance=user_to_store)
        connection.commit()


class SQLAlchemyMessageLogStore(MessageLogStore):
    """
    Message log store for Azure SQL Database

    """
    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="messagelog", **kwargs):
        super().__init__(config, timezone, logger, table_name, **kwargs)
        SQLAlchemyMessageLog.__table__.name = self.table_name

    def get_sqls(self):
        """
        Use api instead.

        """
        return {}

    def prepare_table(self, connection, query_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default None
            Query parameters for checking table
        """
        Base.metadata.create_all(bind=connection.bind, tables=[SQLAlchemyMessageLog.__table__])
        # Always return `False` because the result can't be handled
        return False

    def save(self, request, response, context, connection):
        """
        Write message log

        Parameters
        ----------
        request : Message
            Request message
        response : Response
            Response from chatbot
        context : Context
            Context
        total_ms : int
            Response time (milliseconds)
        connection : Connection
            Connection
        """
        f = self._flatten(request, response, context)
        messagelog = SQLAlchemyMessageLog.from_dict(f)
        messagelog.request_json = request.to_json()
        messagelog.response_json = response.to_json()
        messagelog.context_json = context.to_json()
        connection.add(instance=messagelog)
        connection.commit()


class SQLAlchemyStores(StoreSet):
    connection_provider = SQLAlchemyConnectionProvider
    context_store = SQLAlchemyContextStore
    user_store = SQLAlchemyUserStore
    messagelog_store = SQLAlchemyMessageLogStore



from objson import Serializable


class Group(Serializable):
    """
    Group

    Attributes
    ----------
    id : str
        ID of group
    type : str
        Type of group
    """
    def __init__(self, id=None, type=None):
        """
        Parameters
        ----------
        id : str, default None
            ID of group
        type : str, default None
            Type of group
        """
        self.id = id
        self.type = type



import pyodbc

from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet


class SQLDBConnectionProvider(ConnectionProvider):
    """
    Connection provider for Azure SQL Database

    Attributes
    ----------
    connection_str : str
        Connection string
    """
    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        return pyodbc.connect(self.connection_str)


class SQLDBContextStore(ContextStore):
    """
    Session store for Azure SQL Database to enable successive conversation

    """
    def get_sqls(self):
        """
        Get SQLs used in ContextStore

        Returns
        -------
        sqls : dict
            SQLs used in ContextStore
        """
        return {
            "prepare_check": "select id from dbo.sysobjects where id = object_id('{0}')".format(self.table_name),
            "prepare_create": "create table {0} (channel NVARCHAR(20), channel_user_id NVARCHAR(100), timestamp DATETIME2, topic_name NVARCHAR(100), topic_status NVARCHAR(100), topic_previous NVARCHAR(4000), topic_priority INT, data NVARCHAR(MAX), primary key(channel, channel_user_id))".format(self.table_name),
            "get_context": "select top 1 * from {0} where channel=? and channel_user_id=?".format(self.table_name),
            "save_context": """
                            merge into {0} as A
                            using (select ? as channel, ? as channel_user_id, ? as timestamp, ? as topic_name, ? as topic_status, ? as topic_previous, ? as topic_priority, ? as data) as B
                            on (A.channel = B.channel and A.channel_user_id = B.channel_user_id)
                            when matched then
                            update set timestamp=B.timestamp, topic_name=B.topic_name, topic_status=B.topic_status, topic_previous=B.topic_previous, topic_priority=B.topic_priority, data=B.data
                            when not matched then
                            insert (channel, channel_user_id, timestamp, topic_name, topic_status, topic_previous, topic_priority, data) values (B.channel, B.channel_user_id, B.timestamp, B.topic_name, B.topic_status, B.topic_previous, B.topic_priority, B.data);
                            """.format(self.table_name)
        }


class SQLDBUserStore(UserStore):
    """
    User store for Azure SQL Database

    """
    def get_sqls(self):
        """
        Get SQLs used in UserStore

        Returns
        -------
        sqls : dict
            SQLs used in UserStore
        """
        return {
            "prepare_check": "select id from dbo.sysobjects where id = object_id('{0}')".format(self.table_name),
            "prepare_create": "create table {0} (channel NVARCHAR(20), channel_user_id NVARCHAR(100), user_id NVARCHAR(100), timestamp DATETIME2, name NVARCHAR(100), nickname NVARCHAR(100), profile_image_url NVARCHAR(500), data NVARCHAR(MAX), primary key(channel, channel_user_id))".format(self.table_name),
            "get_user": "select top 1 channel, channel_user_id, user_id, timestamp, name, nickname, profile_image_url, data from {0} where channel=? and channel_user_id=?".format(self.table_name),
            "add_user": "insert into {0} (channel, channel_user_id, user_id, timestamp, name, nickname, profile_image_url, data) values (?,?,?,?,?,?,?,?)".format(self.table_name),
            "save_user": "update {0} set timestamp=?, name=?, nickname=?, profile_image_url=?, data=? where channel=? and channel_user_id=?".format(self.table_name),
        }


class SQLDBMessageLogStore(MessageLogStore):
    """
    Message log store for Azure SQL Database

    """
    def get_sqls(self):
        """
        Get SQLs used in MessageLogStore

        Returns
        -------
        sqls : dict
            SQLs used in MessageLogStore
        """
        return {
            "prepare_check": "select id from dbo.sysobjects where id = object_id('{0}')".format(self.table_name),
            "prepare_create": """
                create table {0} (
                    id INT primary key identity,
                    channel NVARCHAR(20),
                    channel_detail NVARCHAR(100),
                    channel_user_id NVARCHAR(100),
                    request_timestamp DATETIME2,
                    request_id NVARCHAR(100),
                    request_type NVARCHAR(100),
                    request_text NVARCHAR(MAX),
                    request_payloads NVARCHAR(MAX),
                    request_intent NVARCHAR(100),
                    request_is_adhoc BIT,
                    response_type NVARCHAR(100),
                    response_text NVARCHAR(MAX),
                    response_payloads NVARCHAR(MAX),
                    response_milliseconds INT,
                    context_is_new BIT,
                    context_topic_name TEXT,
                    context_topic_status TEXT,
                    context_topic_is_new BIT,
                    context_topic_keep_on BIT,
                    context_topic_priority INT,
                    context_error NVARCHAR(MAX),
                    request_json NVARCHAR(MAX),
                    response_json NVARCHAR(MAX),
                    context_json NVARCHAR(MAX))
                """.format(self.table_name),

            "write": """
                insert into {0} (
                    channel,
                    channel_detail,
                    channel_user_id,
                    request_timestamp,
                    request_id,
                    request_type,
                    request_text,
                    request_payloads,
                    request_intent,
                    request_is_adhoc,
                    response_type,
                    response_text,
                    response_payloads,
                    response_milliseconds,
                    context_is_new,
                    context_topic_name,
                    context_topic_status,
                    context_topic_is_new,
                    context_topic_keep_on,
                    context_topic_priority,
                    context_error,
                    request_json, response_json, context_json)
                values (
                    ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)
                """.format(self.table_name),
        }


class SQLDBStores(StoreSet):
    connection_provider = SQLDBConnectionProvider
    context_store = SQLDBContextStore
    user_store = SQLDBUserStore
    messagelog_store = SQLDBMessageLogStore



import MySQLdb
from MySQLdb.cursors import DictCursor
from MySQLdb.connections import Connection

from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet


class MySQLConnection(Connection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class MySQLConnectionProvider(ConnectionProvider):
    """
    Connection provider for MySQL

    Attributes
    ----------
    connection_str : str
        Connection string
    connection_params : dict
        Parameters for connection
    """
    def __init__(self, connection_str, **kwargs):
        """
        Parameters
        ----------
        connection_str : str
            Connection string
        """
        self.connection_str = connection_str
        self.connection_params = {"cursorclass": DictCursor, "charset": "utf8"}
        param_values = self.connection_str.split(";")
        for pv in param_values:
            if "=" in pv:
                p, v = list(map(str.strip, pv.split("=")))
                self.connection_params[p] = v

    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        return MySQLConnection(**self.connection_params)

    def get_prepare_params(self):
        """
        Get parameters for preparing tables

        Returns
        -------
        prepare_params : tuple or None
            Parameters for preparing tables
        """
        return (self.connection_params["db"], )


class MySQLContextStore(ContextStore):
    def get_sqls(self):
        """
        Get SQLs used in ContextStore

        Returns
        -------
        sqls : dict
            SQLs used in SessionStore
        """
        return {
            "prepare_check": "select * from information_schema.TABLES where TABLE_NAME='{0}' and TABLE_SCHEMA=%s".format(self.table_name),
            "prepare_create": "create table {0} (channel VARCHAR(20), channel_user_id VARCHAR(100), timestamp DATETIME, topic_name VARCHAR(100), topic_status VARCHAR(100), topic_previous VARCHAR(500), topic_priority INT, data JSON, primary key(channel, channel_user_id))".format(self.table_name),
            "get_context": "select channel, channel_user_id, timestamp, topic_name, topic_status, topic_previous, topic_priority, data from {0} where channel=%s and channel_user_id=%s limit 1".format(self.table_name),
            "save_context": "replace into {0} (channel, channel_user_id, timestamp, topic_name, topic_status, topic_previous, topic_priority, data) values (%s,%s,%s,%s,%s,%s,%s,%s)".format(self.table_name),
        }


class MySQLUserStore(UserStore):
    def get_sqls(self):
        """
        Get SQLs used in UserStore

        Returns
        -------
        sqls : dict
            SQLs used in UserRepository
        """
        return {
            "prepare_check": "select * from information_schema.TABLES where TABLE_NAME='{0}' and TABLE_SCHEMA=%s".format(self.table_name),
            "prepare_create": "create table {0} (channel VARCHAR(20), channel_user_id VARCHAR(100), user_id VARCHAR(100), timestamp DATETIME, name VARCHAR(100), nickname VARCHAR(100), profile_image_url VARCHAR(500), data JSON, primary key(channel, channel_user_id))".format(self.table_name),
            "get_user": "select channel, channel_user_id, user_id, timestamp, name, nickname, profile_image_url, data from {0} where channel=%s and channel_user_id=%s limit 1".format(self.table_name),
            "add_user": "insert into {0} (channel, channel_user_id, user_id, timestamp, name, nickname, profile_image_url, data) values (%s,%s,%s,%s,%s,%s,%s,%s)".format(self.table_name),
            "save_user": "update {0} set timestamp=%s, name=%s, nickname=%s, profile_image_url=%s, data=%s where channel=%s and channel_user_id=%s".format(self.table_name),
        }


class MySQLMessageLogStore(MessageLogStore):
    def get_sqls(self):
        """
        Get SQLs used in MessageLogStore

        Returns
        -------
        sqls : dict
            SQLs used in MessageLogger
        """
        return {
            "prepare_check": "select * from information_schema.TABLES where TABLE_NAME='{0}' and TABLE_SCHEMA=%s".format(self.table_name),
            "prepare_create": """
                create table {0} (
                    id INT PRIMARY KEY AUTO_INCREMENT,
                    channel VARCHAR(20),
                    channel_detail VARCHAR(100),
                    channel_user_id VARCHAR(100),
                    request_timestamp DATETIME,
                    request_id VARCHAR(100),
                    request_type VARCHAR(100),
                    request_text VARCHAR(4000),
                    request_payloads JSON,
                    request_intent VARCHAR(100),
                    request_is_adhoc BOOLEAN,
                    response_type VARCHAR(100),
                    response_text VARCHAR(4000),
                    response_payloads JSON,
                    response_milliseconds INT,
                    context_is_new BOOLEAN,
                    context_topic_name TEXT,
                    context_topic_status TEXT,
                    context_topic_is_new BOOLEAN,
                    context_topic_keep_on BOOLEAN,
                    context_topic_priority INT,
                    context_error JSON,
                    request_json JSON,
                    response_json JSON,
                    context_json JSON)
                """.format(self.table_name),
            "write": """
                insert into {0} (
                    channel,
                    channel_detail,
                    channel_user_id,
                    request_timestamp,
                    request_id,
                    request_type,
                    request_text,
                    request_payloads,
                    request_intent,
                    request_is_adhoc,
                    response_type,
                    response_text,
                    response_payloads,
                    response_milliseconds,
                    context_is_new,
                    context_topic_name,
                    context_topic_status,
                    context_topic_is_new,
                    context_topic_keep_on,
                    context_topic_priority,
                    context_error,
                    request_json, response_json, context_json)
                values (
                    %s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
                """.format(self.table_name),
        }


class MySQLStores(StoreSet):
    connection_provider = MySQLConnectionProvider
    context_store = MySQLContextStore
    user_store = MySQLUserStore
    messagelog_store = MySQLMessageLogStore



from .user import User
from .priority import Priority
from .topic import Topic
from .context import Context
from .group import Group
from .payload import Payload
from .message import Message
from .performance import PerformanceInfo
from .response import Response
from .wordnode import WordNode

from .base import JsonSerializable



""" Base class for channel adapters """
from abc import ABC, abstractmethod
import traceback
from logging import Logger
from concurrent.futures import ThreadPoolExecutor

from ..core import Minette


class Adapter(ABC):
    """
    Base class for channel adapters

    Attributes
    ----------
    bot : minette.Minette
        Instance of Minette
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    threads : int
        Number of worker threads to process requests
    executor : ThreadPoolExecutor
        Thread pool of workers
    debug : bool
        Debug mode
    """

    def __init__(self, bot=None, *, threads=None, debug=False, **kwargs):
        """
        Parameters
        ----------
        bot : minette.Minette, default None
            Instance of Minette.
            If None, create new instance of Minette by using `**kwargs`
        threads : int, default None
            Number of worker threads to process requests
        debug : bool, default None
            Debug mode
        """
        self.bot = bot or Minette(**kwargs)
        self.config = self.bot.config
        self.timezone = self.bot.timezone
        self.logger = self.bot.logger
        self.threads = threads
        if self.threads != 0:
            self.logger.info("Use worker threads to handle events")
            self.executor = ThreadPoolExecutor(
                max_workers=self.threads, thread_name_prefix="AdapterThread")
        else:
            self.logger.info("Use main thread to handle events")
            self.executor = None
        self.debug = debug

    def handle_event(self, event):
        """
        Handle event from channel

        Parameters
        ----------
        event : object
            Event data from channel

        Returns
        -------
        channel_messages : list
            List of messages in channel specific format
        """
        if self.debug:
            self.logger.info(event)
        token = self._extract_token(event)
        message = self._to_minette_message(event)
        response = self.bot.chat(message)
        channel_messages = [
            self._to_channel_message(m) for m in response.messages]
        return channel_messages, token

    def _extract_token(self, event):
        """
        Extract token from event

        Parameters
        ----------
        event : object
            Event data from channel

        Returns
        -------
        token : str or object
            Token data for channel
        """
        return ""

    @staticmethod
    @abstractmethod
    def _to_minette_message(event):
        """
        Convert channel event into internal Message object

        Parameters
        ----------
        event : object
            Event data from channel

        Returns
        -------
        message : minette.Message
            Request message
        """
        pass

    @staticmethod
    @abstractmethod
    def _to_channel_message(message):
        """
        Convert internal Message object into channel formatted message

        Parameters
        ----------
        message : minette.Message
            Response message

        Returns
        -------
        channel_message : object
            Channel formatted message
        """
        pass



from datetime import datetime
from pytz import timezone as tz
from copy import copy
from objson import Serializable

from .payload import Payload
from .priority import Priority
from .user import User


class Message(Serializable):
    """
    Message

    Attributes
    ----------
    id : str
        Message ID
    type : str
        Message type
    timestamp : datetime
        Timestamp
    channel : str
        Channel
    channel_detail : str
        Detail information of channel
    channel_user_id : str
        User ID for channel
    channel_message : Any
        Original message from channel
    token : Any
        Token to do actions to the channel
    user : User
        User
    group : Group
        Group
    text : str
        Body of message
    words : list of minette.WordNode
        Word nodes parsed by tagger
    payloads : list of minette.Payload
        Payloads
    intent : str
        Intent extracted from message
    intent_priority : int
        Priority for processing intent
    entities : dict
        Entities extracted from message
    is_adhoc : bool
        Process adhoc or not
    """
    def __init__(self, id=None, type="text", channel="console",
                 channel_detail=None, channel_user_id="anonymous",
                 timestamp=None, channel_message=None, token=None,
                 user=None, group=None, text=None, payloads=None, intent=None,
                 intent_priority=None, entities=None, is_adhoc=False):
        """
        Parameters
        ----------
        id : str, default None
            Message ID
        type : str default "text"
            Message type
        channel : str, default "console"
            Channel
        channel_detail : str, default None
            Detail information of channel
        channel_user_id : str, default "anonymous"
            User ID for channel
        channel_message : Any, default None
            Original message from channel
        token : Any, default None
            Token to do actions to the channel
        text : str, default None
            Body of message
        payloads : list of minette.Payload, default None
            Payloads
        """
        self.id = id or ""
        self.type = type
        self.timestamp = timestamp or datetime.now(tz("UTC"))
        self.channel = channel
        self.channel_detail = channel_detail or ""
        self.channel_user_id = channel_user_id
        self.channel_message = channel_message
        self.token = token or ""
        self.user = user
        self.group = group
        self.text = text or ""
        self.words = []
        self.payloads = payloads or []
        self.intent = intent or ""
        self.intent_priority = intent_priority or Priority.Normal
        self.entities = entities or {}
        self.is_adhoc = is_adhoc

    def to_reply(self, text=None, payloads=None, type="text"):
        """
        Get reply message for this message

        Parameters
        ----------
        text : str, default None
            Body of reply message
        payloads : list of minette.Payload, default None
            Payloads
        type : str default "text"
            Message type

        Returns
        -------
        reply_message : minette.Message
            Reply message for this message
        """
        message = copy(self)
        message.timestamp = datetime.now(message.timestamp.tzinfo)
        message.channel_message = None
        message.type = type
        message.text = text
        message.words = []
        message.payloads = payloads if payloads else []
        message.intent = ""
        message.entities = {}
        message.is_adhoc = False
        return message

    def reply(self, text=None, payloads=None, type="text"):
        print("WARNING: `reply` is deprecated and will be deleted at version 0.5. Use `to_reply` instead.")
        return self.to_reply(text=text, payloads=payloads, type=type)

    def to_dict(self):
        """
        Convert Message object to dict

        Returns
        -------
        message_dict : dict
            Message dictionary
        """
        message_dict = super().to_dict()
        # channel_message is not JSON serializable
        message_dict["channel_message"] = str(message_dict["channel_message"])
        return message_dict

    @classmethod
    def _types(cls):
        return {
            "user": User,
            "payloads": Payload
        }



""" tagger using mecab """
import traceback
import MeCab

from ..models import WordNode
from .base import Tagger


class MeCabNode(WordNode):
    """
    Parsed word node by MeCab

    Attributes
    ----------
    surface : str
        Surface of word
    part : str
        Part of the word
    part_detail1 : str
        Detail1 of part
    part_detail2 : str
        Detail2 of part
    part_detail3 : str
        Detail3 of part
    stem_type : str
        Stem type
    stem_form : str
        Stem form
    word : str
        Word itself
    kana : str
        Japanese kana of the word
    pronunciation : str
        Pronunciation of the word
    """

    @classmethod
    def create(cls, surface, features):
        """
        Create instance of MeCabNode

        Parameters
        ----------
        surface : str
            Surface of the word
        features : list
            Features analyzed by MeCab
        """
        return cls(
            surface=surface,
            part=features[0] if features[0] != "*" else "",
            part_detail1=features[1] if features[1] != "*" else "",
            part_detail2=features[2] if features[2] != "*" else "",
            part_detail3=features[3] if features[3] != "*" else "",
            stem_type=features[4] if features[4] != "*" else "",
            stem_form=features[5] if features[5] != "*" else "",
            word=features[6] if features[6] != "*" else "",
            kana=features[7] if len(features) > 7 else "",
            pronunciation=features[8] if len(features) > 8 else ""
        )


class MeCabTagger(Tagger):
    """
    Tagger using MeCab

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    """

    def parse(self, text):
        """
        Analyze and parse text

        Parameters
        ----------
        text : str
            Text to analyze

        Returns
        -------
        words : list of minette.tagger.mecabtagger.MeCabNode
            MeCab word nodes
        """
        ret = []
        if not text:
            return ret
        try:
            m = MeCab.Tagger("-Ochasen")
            # m.parse("") before m.parseToNode(text) against the bug that node.surface is not set
            m.parse("")
            node = m.parseToNode(text)
            while node:
                features = node.feature.split(",")
                if features[0] != "BOS/EOS":
                    ret.append(MeCabNode.create(node.surface, features))
                node = node.next
        except Exception as ex:
            self.logger.error(
                "MeCab parsing error: "
                + str(ex) + "\n" + traceback.format_exc())
        return ret



""" Core module of minette """
import traceback
import logging
from datetime import datetime
from pytz import timezone as tz
from copy import deepcopy

from .models import (
    Topic,
    Message,
    PerformanceInfo,
    Response
)
from .datastore import (
    ConnectionProvider,
    ContextStore,
    UserStore,
    MessageLogStore,
    SQLiteConnectionProvider,
    SQLiteContextStore,
    SQLiteUserStore,
    SQLiteMessageLogStore
)
from .config import Config
from .dialog import DialogService, DialogRouter
from .tagger import Tagger


class Minette:
    """
    Minette

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : timezone
        Timezone
    logger : logging.Logger
        Logger
    connection_provider : ConnectionProvider
        Connection provider
    context_store: ContextStore
        Context store
    user_store: UserStore
        User store
    messagelog_store: MessageLogStore
        Message store
    default_dialog_service : DialogService
        Dialog service used when intent is not clear
    dialog_router: DialogRouter
        Dialog router to extract intent and entities,
        then returns proper DialogService for intent
    tagger: Tagger
        Morphological analysis engine
    """

    def __init__(self, *, config=None, config_file=None, timezone=None,
                 logger=None, log_file=None, logger_name=None,
                 data_stores=None,
                 connection_provider=None, connection_str=None,
                 context_store=None, context_table=None, context_timeout=None,
                 user_store=None, user_table=None,
                 messagelog_store=None, messagelog_table=None,
                 default_dialog_service=None, dialog_router=None,
                 tagger=None, prepare_table=True, **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        config_file : str, default None
            Path to configuration file. Use `minette.ini` by default.
            config is used when both `config` and `config_file` are passed.
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger. All configuration including filters and handlers
            should be setup before instancing minette.
        log_file : str, default None
            Path to log file. Use `minette.log` by default.
            This is ignored when `logger` is passed.
        logger_name : str, default None
            Name for Logger to separate loggers between bot instances.
            Use `minette` by default.
            This is ignored when `logger` is passed.
        connection_provider : minette.ConnectionProvider or type, default None
            Fully setup instance of `ConnectionProvider` or its class.
            Use `SQLiteConnectionProvider` by default.
        connection_str : str, default None
            Connection string to create instance of ConnectionProvider.
            This is used when the class of ConnectionProvider
            passed for `connection_provider`.
        context_store: minette.ContextStore or type, default None
            Fully setup instance of `ContextStore` or its class.
            Use `SQLiteContextStore` by default.
        context_table: str, default None
            Database table name for ContextStore. Use `context` by default.
            This is ignored when instance of `ContextStore` is passed as
            `context_store`.
        context_timeout: int, default None
            Timeout of context(seconds, default 300).
        user_store: minette.UserStore or type, default None
            Fully setup instance of `UserStore` or its class.
            Use `SQLiteUserStore` by default.
        user_table: str, default None
            Database table name for UserStore. Use `user` by default.
            This is ignored when instance of `UserStore` is passed as
            `user_store`.
        messagelog_store: minette.MessageLogStore or type, default None
            Fully setup instance of `MessageLogStore` or its class.
            Use `SQLiteMessageLogStore` by default.
        messagelog_table: str, default None
            Database table name for MessageLogStore. Use `messegelog`
            by default.
            This is ignored when instance of `MessageLogStore` is passed as
            `messagelog_store`.
        default_dialog_service : minette.DialogService or type, default None
            Dialog service used when intent is not clear.
        dialog_router: minette.DialogRouter or type, default None
            Dialog router to extract intent and entities,
            and return proper DialogService for intent
        tagger: minette.Tagger or type, default None
            Morphological analysis engine
        prepare_table: bool, default True
            Create tables for data stores if they don't exist.
        """
        # setup essensial members for other members
        if config:
            self.config = config
        else:
            self.config = Config(config_file or "minette.ini")
        self.timezone = timezone or tz(self.config.get("timezone") or "UTC")
        self.logger = self._get_logger(
            logger, log_file=log_file, logger_name=logger_name)
        self.connection_provider = self._get_connection_provider(
            connection_provider or (
                data_stores.connection_provider if data_stores else None),
            connection_str=connection_str, **kwargs)

        # make arguments dict
        setter_args = {
            "config": self.config,
            "timezone": self.timezone,
            "logger": self.logger,
            "connection_provider": self.connection_provider,
            "context_store": context_store or (
                data_stores.context_store if data_stores else None),
            "context_table": context_table,
            "context_timeout": context_timeout,
            "user_store": user_store or (
                data_stores.user_store if data_stores else None),
            "user_table": user_table,
            "messagelog_store": messagelog_store or (
                data_stores.messagelog_store if data_stores else None),
            "messagelog_table": messagelog_table,
            "dialog_router": dialog_router,
            "default_dialog_service": default_dialog_service,
            "tagger": tagger,
        }
        setter_args.update({k: v for k, v in kwargs.items() if k not in setter_args})

        # setup members
        self.context_store = self._get_context_store(**setter_args)
        self.user_store = self._get_user_store(**setter_args)
        self.messagelog_store = self._get_messagelog_store(**setter_args)
        self.default_dialog_service = default_dialog_service
        self.dialog_router = self._get_dialog_router(**setter_args)
        self.tagger = self._get_tagger(**setter_args)

        # prepare tables
        if prepare_table is True:
            connection = self.connection_provider.get_connection()
            prepare_params = self.connection_provider.get_prepare_params()
            self.context_store.prepare_table(connection, prepare_params)
            self.user_store.prepare_table(connection, prepare_params)
            self.messagelog_store.prepare_table(connection, prepare_params)
            if hasattr(connection, "close"):
                connection.close()

    def _get_logger(self, logger, log_file=None, logger_name=None):
        lg = logger
        # use passed logger if already setup
        if lg:
            return lg
        lg = logging.getLogger(
            logger_name or self.config.get("logger_name") or "minette")
        # use logger if handlers are already setup
        # evaluate len() because lg.hasHandlers() returns True
        # even when lg.handlers == [] in some cases(e.g. pytest)
        if len(lg.handlers) > 0:
            return lg
        # setup logger
        lg.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        stream_handler.setFormatter(formatter)
        lg.addHandler(stream_handler)
        file_handler = logging.FileHandler(
            filename=log_file or
            self.config.get("log_file") or "minette.log"
        )
        file_handler.setLevel(logging.ERROR)
        file_handler.setFormatter(formatter)
        lg.addHandler(file_handler)
        return lg

    def _get_connection_provider(self, connection_provider,
                                 connection_str=None, **kwargs):
        cp = connection_provider or SQLiteConnectionProvider
        if issubclass(cp, ConnectionProvider):
            cp = cp(
                connection_str=connection_str or
                self.config.get("connection_str") or "minette.db",
                **kwargs
            )
        return cp

    def _get_context_store(self, context_store, context_table=None,
                           context_timeout=None, **kwargs):
        ss = context_store or SQLiteContextStore
        if issubclass(ss, ContextStore):
            ss = ss(
                table_name=context_table or
                self.config.get("context_table") or "context",
                timeout=context_timeout or
                self.config.get("context_timeout") or 300,
                **kwargs
            )
        return ss

    def _get_user_store(self, user_store, user_table=None, **kwargs):
        us = user_store or SQLiteUserStore
        if issubclass(us, UserStore):
            us = us(
                table_name=user_table or
                self.config.get("user_table") or "user",
                **kwargs
            )
        return us

    def _get_messagelog_store(self, messagelog_store, messagelog_table=None,
                              **kwargs):
        ms = messagelog_store or SQLiteMessageLogStore
        if issubclass(ms, MessageLogStore):
            ms = ms(
                table_name=messagelog_table or
                self.config.get("messagelog_table") or "messagelog",
                **kwargs
            )
        return ms

    def _get_dialog_router(self, dialog_router, default_dialog_service=None,
                           **kwargs):
        dr = dialog_router or DialogRouter
        if issubclass(dr, DialogRouter):
            dr = dr(default_dialog_service=default_dialog_service, **kwargs)
        return dr

    def _get_tagger(self, tagger, **kwargs):
        tg = tagger or Tagger
        if issubclass(tg, Tagger):
            tg = tg(**kwargs)
        return tg

    def chat(self, request):
        """
        Get response from chatbot

        Examples
        --------
        >>> from minette import *
        >>> bot = Minette(defautl_dialog_service=EchoDialogService)
        >>> response = bot.chat("hello")
        >>> response.messages[0].text
        "You said: hello"

        Parameters
        ----------
        request : minette.Message or str
            Message to chatbot

        Returns
        -------
        response : minette.Response
            Response from chatbot
        """
        connection = None
        try:
            performance = PerformanceInfo()
            if isinstance(request, str):
                request = Message(text=request, timestamp=datetime.now(self.timezone))
            # connection
            connection = self.connection_provider.get_connection()
            performance.append("connection_provider.get_connection")
            # tagger
            request.words = self.tagger.parse(request.text)
            performance.append("tagger.parse")
            # user
            request.user = self._get_user(request, connection)
            performance.append("get_user")
            # context
            context = self._get_context(request, connection)
            performance.append("get_context")
            # route dialog
            dialog_service = self.dialog_router.execute(
                request=request, context=context,
                connection=connection, performance=performance)
            performance.append("dialog_router.execute")
            # process dialog
            response = dialog_service.execute(
                request=request, context=context,
                connection=connection, performance=performance)
            performance.append("dialog_service.execute")
            # save context
            context = self._save_context(context, connection)
            performance.append("save_context")
            # save user
            self._save_user(request.user, connection)
            performance.append("save_user")
        except Exception as ex:
            self.logger.error(
                "Error occured in chat: "
                + str(ex) + "\n" + traceback.format_exc())
            response = Response()
        finally:
            # set performance info to response
            response.performance = performance
            if connection:
                # message log
                try:
                    self.messagelog_store.save(
                        request, response, context, connection)
                except Exception as ex:
                    self.logger.error(
                        "Error occured in logging message: "
                        + str(ex) + "\n" + traceback.format_exc())
                # close connection
                if hasattr(connection, "close"):
                    connection.close()
        return response

    def _get_user(self, request, connection):
        user_scope = request.channel
        if self.config.get("user_scope") == "channel_detail":
            user_scope += "_" + request.channel_detail
        return self.user_store.get(
            user_scope, request.channel_user_id, connection)

    def _save_user(self, user, connection):
        self.user_store.save(user, connection)

    def _get_context(self, request, connection):
        context_scope = request.channel
        if self.config.get("context_scope") == "channel_detail":
            context_scope += "_" + request.channel_detail
        if request.group:
            context = self.context_store.get(
                context_scope, request.group.id, connection)
        else:
            context = self.context_store.get(
                context_scope, request.channel_user_id, connection)
        return context

    def _save_context(self, context, connection):
        context_for_log = deepcopy(context)
        context.reset(self.config.get("keep_context_data", False))
        self.context_store.save(context, connection)
        return context_for_log



import requests
from objson import Serializable


class Payload(Serializable):
    """
    Payload

    Attributes
    ----------
    content_type : str
        Content type of payload
    url : str
        Url to get content
    thumb : str
        URL to get thumbnail image
    headers : dict
        HTTP Headers required for getting content
    content : Any
        Content itself
    """
    def __init__(self, content_type="image", url=None, thumb=None,
                 headers=None, content=None):
        """
        Parameters
        ----------
        content_type : str, default "image"
            Content type of payload
        url : str, default None
            Url to get content
        thumb : str, default None
            URL to get thumbnail image
        headers : dict, default None
            HTTP Headers required for getting content
        content : Any, default None
            Content itself
        """
        self.content_type = content_type
        self.url = url
        self.thumb = thumb if thumb is not None else url
        self.headers = headers or {}
        self.content = content

    def get(self, set_content=False):
        """
        Get content data from url

        Parameters
        ----------
        set_content : bool, default False
            If True, set content data to content property

        Returns
        -------
        content : Any
            Content
        """
        data = requests.get(self.url, headers=self.headers, timeout=60).content
        if set_content:
            self.content = data
        return data

    def save(self, filepath):
        """
        Save content data to file

        Parameters
        ----------
        filepath : str
            File path to save content
        """
        data = self.get()
        with open(filepath, "wb") as f:
            f.write(data)



""" Scheduler for periodic tasks """
import traceback
from logging import getLogger
import schedule
import time
from concurrent.futures import ThreadPoolExecutor


class Task:
    """
    Base class of tasks

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    connection_provider : minette.ConnectionProvider
        Connection provider to use database in each tasks
    """
    def __init__(self, config=None, timezone=None, logger=None,
                 connection_provider=None, **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        connection_provider : minette.ConnectionProvider
            Connection provider to use database in each tasks
        """
        self.config = config
        self.timezone = timezone
        self.logger = logger or getLogger(__name__)
        self.connection_provider = connection_provider

    def do(self, **kwargs):
        """
        Implement your periodic task

        """
        self.logger.error("Task is not implemented")


class Scheduler:
    """
    Task scheduler for periodic tasks

    Examples
    --------
    To start doing scheduled tasks, just create `Scheduler` instance
    and register task(s), then call `start()`

    >>> my_scheduler = MyScheduler()
    >>> my_scheduler.every_minutes(MyTask)
    >>> my_scheduler.start()

    To register tasks, this class provides shortcut methods.
    Each tasks run at worker threads.

    >>> my_scheduler.every_minutes(MyTask)
    >>> my_scheduler.every_seconds(MyTask, seconds=5)
    >>> my_scheduler.every_seconds(MyTask, seconds=5, arg1="val1", arg2="val2")

    You can also use internal `schedule` to register tasks
    then the tasks run at main thread.

    >>> my_scheduler.schedule.every().minute.do(self.create_task(MyTask))
    >>> my_scheduler.schedule.every().hour.do(self.create_task(YourTask))

    Notes
    -----
    How to execute jobs in parallel?
    https://schedule.readthedocs.io/en/stable/faq.html#how-to-execute-jobs-in-parallel

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    threads : int
        Number of worker threads to process tasks
    connection_provider : minette.ConnectionProvider
        Connection provider to use database in each tasks
    schedule : schedule
        schedule module
    executor : concurrent.futures.ThreadPoolExecutor
        Executor to run tasks at worker threads
    """

    def __init__(self, config=None, timezone=None, logger=None, threads=None,
                 connection_provider=None, **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        threads : int, default None
            Number of worker threads to process tasks
        connection_provider : ConnectionProvider, default None
            Connection provider to use database in each tasks
        """
        self.config = config
        self.timezone = timezone
        self.logger = logger or getLogger(__name__)
        self.threads = threads
        self.connection_provider = connection_provider
        self.schedule = schedule
        self.executor = ThreadPoolExecutor(
            max_workers=self.threads, thread_name_prefix="SchedulerThread")
        self._is_running = False

    @property
    def is_running(self):
        return self._is_running

    def create_task(self, task_class, **kwargs):
        """
        Create and return callable function of task

        Parameters
        ----------
        task_class : type
            Class of task

        Returns
        -------
        task_method : callable
            Callable interface of task
        """
        if isinstance(task_class, type):
            if issubclass(task_class, Task):
                return task_class(
                    config=self.config,
                    timezone=self.timezone,
                    logger=self.logger,
                    connection_provider=self.connection_provider,
                    **kwargs).do
            else:
                raise TypeError(
                    "task_class should be a subclass of minette.Task " +
                    "or callable, not {}".format(task_class.__name__))

        elif callable(task_class):
            return task_class

        else:
            raise TypeError(
                "task_class should be a subclass of minette.Task " +
                "or callable, not the instance of {}".format(
                    task_class.__class__.__name__))

    def every_seconds(self, task, seconds=1, *args, **kwargs):
        self.schedule.every(seconds).seconds.do(
            self.executor.submit, self.create_task(task), *args, **kwargs)

    def every_minutes(self, task, minutes=1, *args, **kwargs):
        self.schedule.every(minutes).minutes.do(
            self.executor.submit, self.create_task(task), *args, **kwargs)

    def every_hours(self, task, hours=1, *args, **kwargs):
        self.schedule.every(hours).hours.do(
            self.executor.submit, self.create_task(task), *args, **kwargs)

    def every_days(self, task, days=1, *args, **kwargs):
        self.schedule.every(days).days.do(
            self.executor.submit, self.create_task(task), *args, **kwargs)

    def start(self):
        """
        Start scheduler
        """
        self.logger.info("Task scheduler started")
        self._is_running = True
        while self._is_running:
            self.schedule.run_pending()
            time.sleep(1)
        self.logger.info("Task scheduler stopped")

    def stop(self):
        """
        Stop scheduler
        """
        self._is_running = False



""" tagger using janome """
import traceback
from janome.tokenizer import Tokenizer

from ..models import WordNode
from .base import Tagger


class JanomeNode(WordNode):
    """
    Parsed word node by Janome

    Attributes
    ----------
    surface : str
        Surface of word
    part : str
        Part of the word
    part_detail1 : str
        Detail1 of part
    part_detail2 : str
        Detail2 of part
    part_detail3 : str
        Detail3 of part
    stem_type : str
        Stem type
    stem_form : str
        Stem form
    word : str
        Word itself
    kana : str
        Japanese kana of the word
    pronunciation : str
        Pronunciation of the word
    """

    @classmethod
    def create(cls, surface, features):
        """
        Create instance of JanomeNode

        Parameters
        ----------
        surface : str
            Surface of the word
        features : janome.Token
            Features analyzed by Janome
        """
        ps = features.part_of_speech.split(",")
        return cls(
            surface=surface,
            part=ps[0] if len(ps) > 1 and ps[0] != "*" else "",
            part_detail1=ps[1] if len(ps) > 2 and ps[1] != "*" else "",
            part_detail2=ps[2] if len(ps) > 3 and ps[2] != "*" else "",
            part_detail3=ps[3] if len(ps) > 4 and ps[3] != "*" else "",
            stem_type=features.infl_type if features.infl_type != "*" else "",
            stem_form=features.infl_form if features.infl_form != "*" else "",
            word=features.base_form if features.base_form != "*" else "",
            kana=features.reading if features.reading != "*" else "",
            pronunciation=features.phonetic if features.phonetic != "*" else ""
        )


class JanomeTagger(Tagger):
    """
    Tagger using Janome

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    """

    def __init__(self, config=None, timezone=None, logger=None, *,
                 user_dic=None, **kwargs):
        """
        Parameters
        ----------
        config : Config, default None
            Configuration
        timezone : timezone, default None
            Timezone
        logger : Logger, default None
            Logger
        user_dic : str, default None
             Path to user dictionary (MeCab IPADIC format)
        """
        super().__init__(logger=logger, config=config, timezone=timezone)
        self.user_dic = user_dic or config.get("janome_userdic") if config else None
        if self.user_dic:
            self.tokenizer = Tokenizer(self.user_dic, udic_enc="utf8")
        else:
            self.tokenizer = Tokenizer()

    def parse(self, text):
        """
        Parse and annotate using Janome

        Parameters
        ----------
        text : str
            Text to analyze

        Returns
        -------
        words : list of minette.minette.tagger.janometagger.JanomeNode
            Janome nodes
        """
        ret = []
        if not text:
            return ret
        try:
            for token in self.tokenizer.tokenize(text):
                ret.append(JanomeNode.create(token.surface, token))
        except Exception as ex:
            self.logger.error(
                "Janome parsing error: "
                + str(ex) + "\n" + traceback.format_exc())
        return ret



""" Base class for MessageLogStore """
from abc import ABC, abstractmethod
from logging import getLogger
from pytz import timezone as tz

import objson


class MessageLogStore(ABC):
    """
    Base class for MessageLogStore to analyze the communication

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    table_name : str
        Database table name for read/write message log data
    sqls : dict
        SQLs used in ContextStore
    """
    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="messagelog", **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        table_name : str, default "messagelog"
            Database table name for read/write message log data
        """
        self.config = config
        self.timezone = timezone or (
            tz(config.get("timezone", default="UTC")) if config else tz("UTC"))
        self.logger = logger if logger else getLogger(__name__)
        self.table_name = table_name
        self.sqls = self.get_sqls()

    @abstractmethod
    def get_sqls(self):
        pass

    def prepare_table(self, connection, prepare_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default tuple()
            Query parameters for checking table
        """
        cursor = connection.cursor()
        cursor.execute(self.sqls["prepare_check"], prepare_params or tuple())
        if not cursor.fetchone():
            cursor.execute(self.sqls["prepare_create"])
            connection.commit()
            return True
        else:
            return False

    def _flatten(self, request, response, context):
        return {
            # request
            "channel": request.channel,
            "channel_detail": request.channel_detail,
            "channel_user_id": request.channel_user_id,
            "request_timestamp": request.timestamp,
            "request_id": request.id,
            "request_type": request.type,
            "request_text": request.text,
            "request_payloads": objson.dumps(
                [p.to_dict() for p in request.payloads]),
            "request_intent": request.intent,
            "request_is_adhoc": request.is_adhoc,
            # response
            "response_type": response.messages[0].type if response.messages else "",
            "response_text": response.messages[0].text if response.messages else "",
            "response_payloads": objson.dumps([p.to_dict() for p in response.messages[0].payloads]) if response.messages else "",
            "response_milliseconds": response.performance.milliseconds,
            # context
            "context_is_new": context.is_new,
            "context_topic_name": context.topic.name,
            "context_topic_status": context.topic.status,
            "context_topic_is_new": context.topic.is_new,
            "context_topic_keep_on": context.topic.keep_on,
            "context_topic_priority": context.topic.priority,
            "context_error": objson.dumps(context.error),
        }

    def save(self, request, response, context, connection):
        """
        Write message log

        Parameters
        ----------
        request : minette.Message
            Request to chatbot
        response : minette.Response
            Response from chatbot
        context : minette.Context
            Context
        connection : Connection
            Connection
        """
        f = self._flatten(request, response, context)
        cursor = connection.cursor()
        cursor.execute(self.sqls["write"], (
            f["channel"],
            f["channel_detail"],
            f["channel_user_id"],
            f["request_timestamp"],
            f["request_id"],
            f["request_type"],
            f["request_text"],
            f["request_payloads"],
            f["request_intent"],
            f["request_is_adhoc"],
            f["response_type"],
            f["response_text"],
            f["response_payloads"],
            f["response_milliseconds"],
            f["context_is_new"],
            f["context_topic_name"],
            f["context_topic_status"],
            f["context_topic_is_new"],
            f["context_topic_keep_on"],
            f["context_topic_priority"],
            f["context_error"],
            request.to_json(),
            response.to_json(),
            context.to_json())
        )
        connection.commit()



""" Configuration management """
import os
from configparser import ConfigParser


class Config:
    """
    Configuration management

    Attributes
    ----------
    confg_parser : ConfigParser
        ConfigParser used internally
    """

    def __init__(self, config_file):
        """
        Parameters
        ----------
        config_file : str
            Path to configuration file
        """
        self.confg_parser = ConfigParser()
        if config_file and len(self.confg_parser.read(config_file)) == 0:
            print("Can't read/find configuration file: {}".format(config_file))
            print("Initialize with default configuration instead")

        if not self.confg_parser.has_section("minette"):
            self.confg_parser.add_section("minette")
            self.confg_parser.set("minette", "timezone", "UTC")
            self.confg_parser.set(
                "minette", "log_file", "ENV::MINETTE_LOGFILE")
            self.confg_parser.set(
                "minette", "connection_str", "ENV::MINETTE_CONNECTION_STR")

    def get(self, key, default=None, section="minette"):
        if section in self.confg_parser.sections():
            ret = self.confg_parser[section].get(key, default)
        else:
            ret = default
        if str(ret).startswith("ENV::"):
            ret = os.environ.get(ret[5:], default)
        return ret



from objson import Serializable

from .performance import PerformanceInfo
from .message import Message


class Response(Serializable):
    """
    Response from chatbot

    Attributes
    ----------
    messages : list of minette.Message
        Response messages
    headers : dict
        Response header
    performance : minette.PerformanceInfo
        Performance information of each steps in chat()
    """
    def __init__(self, messages=None, headers=None, performance=None):
        """
        Parameters
        ----------
        messages : list of minette.Message, default None
            Response messages. If None, `[]` is set to `messages`.
        headers : dict, default None
            Response headers. If None, `{}` is set to `headers`
        performance : minette.PerformanceInfo, default None
            Performance information of each steps in chat().
            If None, create new PerformanceInfo object.
        """
        self.messages = messages or []
        self.headers = headers or {}
        self.performance = performance or PerformanceInfo()

    @classmethod
    def _types(cls):
        return {
            "messages": Message,
            "performance": PerformanceInfo
        }



from time import time

from ..core import Minette
from ..models import Message


class MinetteForTest(Minette):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.default_channel = kwargs.get("default_channel", "")
        self.case_id = str(int(time() * 10000000))

    def chat(self, request, **kwargs):
        self.logger.info("start testcase: " + self.case_id)
        # convert to Message
        if isinstance(request, str):
            request = Message(text=request, **kwargs)
        # set channel and channel_user_id for this test case
        if request.channel == "console":
            request.channel = self.default_channel or request.channel
        if request.channel_user_id == "anonymous":
            request.channel_user_id = "user" + self.case_id
        # chat and return response
        response = super().chat(request)
        if response.messages:
            response.text = response.messages[0].text
        else:
            response.text = ""
        self.logger.info("end testcase: " + self.case_id)
        return response



from objson import Serializable

from .priority import Priority


class Topic(Serializable):
    """
    Topic

    Attributes
    ----------
    name : str
        Name of topic
    status : str
        Status of topic
    is_new : bool
        True if topic starts at this turn
    keep_on : bool
        True to keep this topic at next turn
    previous : minette.Topic
        Previous topic object
    priority : int
        Priority of topic
    is_changed : bool
        True if topic is changed at this turn
    """
    def __init__(self):
        self.name = ""
        self.status = ""
        self.is_new = False
        self.keep_on = False
        self.previous = None
        self.priority = Priority.Normal

    @property
    def is_changed(self):
        if self.previous and self.previous.name == self.name:
            return False
        else:
            return True



""" Base class for ConnectionProvider """
from abc import ABC, abstractmethod


class ConnectionProvider(ABC):
    """
    Base class for ConnectionProvider

    Attributes
    ----------
    connection_str : str
        Connection string
    """

    def __init__(self, connection_str, **kwargs):
        """
        Parameters
        ----------
        connection_str : str
            Connection string
        """
        self.connection_str = connection_str

    @abstractmethod
    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        pass

    def get_prepare_params(self):
        """
        Get parameters for preparing tables

        Returns
        -------
        prepare_params : tuple or None
            Parameters for preparing tables
        """
        return None



""" Set of data stores and connection provider using SQLite """
import sqlite3

from .connectionprovider import ConnectionProvider
from .contextstore import ContextStore
from .userstore import UserStore
from .messagelogstore import MessageLogStore
from .storeset import StoreSet


class SQLiteConnectionProvider(ConnectionProvider):
    """
    Connection provider for SQLite

    Attributes
    ----------
    connection_str : str
        Connection string
    """
    def get_connection(self):
        """
        Get connection

        Returns
        -------
        connection : Connection
            Database connection
        """
        connection = sqlite3.connect(
            self.connection_str, detect_types=sqlite3.PARSE_DECLTYPES)
        connection.row_factory = sqlite3.Row
        return connection


class SQLiteContextStore(ContextStore):
    """
    ContextStore using SQLite

    """
    def get_sqls(self):
        """
        Get SQLs used in ContextStore

        Returns
        -------
        sqls : dict
            SQLs used in ContextStore
        """
        return {
            "prepare_check": """
                select * from sqlite_master where type='table' and name='{0}'
                """.format(self.table_name),
            "prepare_create": """
                create table {0} (
                    channel TEXT, channel_user_id TEXT, timestamp TIMESTAMP,
                    topic_name TEXT, topic_status TEXT, topic_previous TEXT,
                    topic_priority INTEGER, data TEXT, primary key(channel,
                    channel_user_id))""".format(self.table_name),
            "get_context": """
                select
                    channel, channel_user_id, timestamp, topic_name,
                    topic_status, topic_previous, topic_priority, data
                from {0}
                where
                    channel=? and channel_user_id=? limit 1
                """.format(self.table_name),
            "save_context": """
                replace into {0} (
                    channel, channel_user_id, timestamp, topic_name,
                    topic_status, topic_previous, topic_priority, data)
                values (
                    ?,?,?,?,?,?,?,?)
                """.format(self.table_name),
        }


class SQLiteUserStore(UserStore):
    """
    UserStore using SQLite

    """
    def get_sqls(self):
        """
        Get SQLs used in UserStore

        Returns
        -------
        sqls : dict
            SQLs used in UserStore
        """
        return {
            "prepare_check": """
                select * from sqlite_master where type='table' and name='{0}'
                """.format(self.table_name),
            "prepare_create": """
                create table {0} (
                    channel TEXT, channel_user_id TEXT, user_id TEXT,
                    timestamp TIMESTAMP, name TEXT, nickname TEXT,
                    profile_image_url TEXT, data TEXT,
                    primary key(channel, channel_user_id))
                """.format(self.table_name),
            "get_user": """
                select
                    channel, channel_user_id, user_id,timestamp, name,
                    nickname, profile_image_url, data
                from {0}
                where
                    channel=? and channel_user_id=? limit 1
                """.format(self.table_name),
            "add_user": """
                insert into {0} (
                    channel, channel_user_id, user_id, timestamp, name,
                    nickname, profile_image_url, data)
                values (
                    ?,?,?,?,?,?,?,?)
                """.format(self.table_name),
            "save_user": """
                update {0}
                set
                    timestamp=?, name=?, nickname=?, profile_image_url=?,
                    data=?
                where
                    channel=? and channel_user_id=?
                """.format(self.table_name),
        }


class SQLiteMessageLogStore(MessageLogStore):
    """
    MessageLogStore using SQLite

    """
    def get_sqls(self):
        """
        Get SQLs used in MessageLogStore

        Returns
        -------
        sqls : dict
            SQLs used in MessageLogStore
        """
        return {
            "prepare_check": """
                select * from sqlite_master where type='table' and name='{0}'
                """.format(self.table_name),
            "prepare_create": """
                create table {0} (
                    id INTEGER PRIMARY KEY,
                    channel TEXT,
                    channel_detail TEXT,
                    channel_user_id TEXT,
                    request_timestamp TIMESTAMP,
                    request_id TEXT,
                    request_type TEXT,
                    request_text TEXT,
                    request_payloads TEXT,
                    request_intent TEXT,
                    request_is_adhoc BOOLEAN,
                    response_type TEXT,
                    response_text TEXT,
                    response_payloads TEXT,
                    response_milliseconds INT,
                    context_is_new BOOLEAN,
                    context_topic_name TEXT,
                    context_topic_status TEXT,
                    context_topic_is_new BOOLEAN,
                    context_topic_keep_on BOOLEAN,
                    context_topic_priority INTEGER,
                    context_error TEXT,
                    request_json TEXT,
                    response_json TEXT,
                    context_json TEXT)
                """.format(self.table_name),
            "write": """
                insert into {0} (
                    channel,
                    channel_detail,
                    channel_user_id,
                    request_timestamp,
                    request_id,
                    request_type,
                    request_text,
                    request_payloads,
                    request_intent,
                    request_is_adhoc,
                    response_type,
                    response_text,
                    response_payloads,
                    response_milliseconds,
                    context_is_new,
                    context_topic_name,
                    context_topic_status,
                    context_topic_is_new,
                    context_topic_keep_on,
                    context_topic_priority,
                    context_error,
                    request_json, response_json, context_json)
                values (
                    ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)
                """.format(self.table_name),
        }


class SQLiteStores(StoreSet):
    """
    Set of data stores and connection provider using SQLite

    """
    connection_provider = SQLiteConnectionProvider
    context_store = SQLiteContextStore
    user_store = SQLiteUserStore
    messagelog_store = SQLiteMessageLogStore



from time import time
from objson import Serializable


class PerformanceInfo(Serializable):
    """
    Performance information of each steps

    Attributes
    ----------
    start_time : float
        Unix epoch seconds at start
    ticks : list
        Seconds since start_time
    milliseconds : int
        Total processing time in milliseconds
    """
    def __init__(self):
        self.start_time = time()
        self.ticks = []
        self.milliseconds = 0

    def append(self, comment):
        """
        Append current performance timestamp

        Parameters
        ----------
        comment : str
            Comment to identify each steps
        """
        self.ticks.append((comment, time() - self.start_time))
        self.milliseconds = int(self.ticks[-1][1] * 1000)



from abc import ABC, abstractmethod
from objson import Serializable


class WordNode(ABC, Serializable):
    """
    Base class of parsed word

    Attributes
    ----------
    surface : str
        Surface of word
    part : str
        Part of the word
    part_detail1 : str
        Detail1 of part
    part_detail2 : str
        Detail2 of part
    part_detail3 : str
        Detail3 of part
    stem_type : str
        Stem type
    stem_form : str
        Stem form
    word : str
        Word itself
    kana : str
        Japanese kana of the word
    pronunciation : str
        Pronunciation of the word
    """
    def __init__(self, surface, part, part_detail1, part_detail2, part_detail3,
                 stem_type, stem_form, word, kana, pronunciation):
        self.surface = surface
        self.part = part
        self.part_detail1 = part_detail1
        self.part_detail2 = part_detail2
        self.part_detail3 = part_detail3
        self.stem_type = stem_type
        self.stem_form = stem_form
        self.word = word
        self.kana = kana
        self.pronunciation = pronunciation

    @classmethod
    @abstractmethod
    def create(cls, surface, features):
        pass



from .version import __version__

from .config import Config
from .core import Minette
from .datastore import (
    ConnectionProvider,
    ContextStore,
    UserStore,
    MessageLogStore,
    StoreSet,
    SQLiteConnectionProvider,
    SQLiteContextStore,
    SQLiteUserStore,
    SQLiteMessageLogStore,
    SQLiteStores
)
from .dialog import (
    DialogService,
    EchoDialogService,
    ErrorDialogService,
    DialogRouter
)
from .models import *
from .tagger import (
    Tagger,
    MeCabServiceTagger,
    MeCabServiceNode
)
from .adapter import Adapter
from .scheduler import Task, Scheduler



from setuptools import setup, find_packages

with open("./minette/version.py") as f:
    exec(f.read())

setup(
    name="minette",
    version=__version__,
    url="https://github.com/uezo/minette-python",
    author="uezo",
    author_email="uezo@uezo.net",
    maintainer="uezo",
    maintainer_email="uezo@uezo.net",
    description="Minette is a minimal and extensible chatbot framework. It is extremely easy to develop and the architecture preventing to be spaghetti code enables you to scale up to complex chatbot.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["examples*", "develop*", "tests*"]),
    install_requires=["requests", "pytz", "schedule", "objson-py"],
    license="Apache v2",
    classifiers=[
        "Programming Language :: Python :: 3"
    ]
)



import traceback
from datetime import datetime

from linebot import LineBotApi, WebhookParser
from linebot.exceptions import InvalidSignatureError
from linebot.models import (
    Event, BeaconEvent, FollowEvent, JoinEvent, LeaveEvent, MessageEvent,
    PostbackEvent, UnfollowEvent, MemberJoinedEvent, MemberLeftEvent,
    TextMessage, ImageMessage, AudioMessage, VideoMessage, LocationMessage,
    StickerMessage, TextSendMessage, ImageSendMessage, AudioSendMessage,
    VideoSendMessage, LocationSendMessage, StickerSendMessage,
    ImagemapSendMessage, TemplateSendMessage, FlexSendMessage
)

from .base import Adapter
from ..models import (
    Message,
    Payload,
    Group,
    Response,
    Context
)


class LineAdapter(Adapter):
    """
    Adapter for LINE Messaging API

    Attributes
    ----------
    bot : minette.Minette
        Instance of Minette
    channel_secret : str
        Channel Secret
    channel_access_token : str
        Channel Access Token
    parser : WebhookParser
        WebhookParser
    api : LineBotApi
        LineBotApi
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    threads : int
        Number of worker threads to process requests
    executor : ThreadPoolExecutor
        Thread pool of workers
    debug : bool
        Debug mode
    """

    def __init__(self, bot=None, *, threads=None, debug=False,
                 channel_secret=None, channel_access_token=None, **kwargs):
        """
        Parameters
        ----------
        bot : minette.Minette, default None
            Instance of Minette.
            If None, create new instance of Minette by using `**kwargs`
        channel_secret : str, default None
            Channel Secret
        channel_access_token : str, default None
            Channel Access Token
        threads : int, default None
            Number of worker threads to process requests
        debug : bool, default None
            Debug mode
        """
        super().__init__(bot=bot, threads=threads, debug=debug, **kwargs)
        self.channel_secret = channel_secret or \
            self.config.get(section="line_bot_api", key="channel_secret")
        self.channel_access_token = channel_access_token or \
            self.config.get(section="line_bot_api", key="channel_access_token")
        self.parser = WebhookParser(self.channel_secret)
        self.api = LineBotApi(self.channel_access_token)

    def handle_http_request(self, request_data, request_headers):
        """
        Interface to chat with LINE Bot

        Parameters
        ----------
        request_data : list of byte
            Request data from LINE Messaging API as string
        request_headers : dict
            Request headers from LINE Messaging API as dict

        Returns
        -------
        response : Response
            Response that shows queued status
        """
        try:
            events = self.parser.parse(
                request_data.decode("utf-8"),
                request_headers.get("X-Line-Signature", ""))
            for ev in events:
                if self.executor:
                    self.executor.submit(self.handle_event, ev)
                else:
                    self.handle_event(ev)
            return Response(messages=[Message(text="done", type="system")])
        except InvalidSignatureError as ise:
            self.logger.error(
                "Request signiture is invalid: "
                + str(ise) + "\n" + traceback.format_exc())
            return Response(
                messages=[Message(text="invalid signiture", type="system")])
        except Exception as ex:
            self.logger.error(
                "Request parsing error: "
                + str(ex) + "\n" + traceback.format_exc())
            return Response(
                messages=[Message(text="failure in parsing request",
                                  type="system")])

    def handle_event(self, event):
        channel_messages, token = super().handle_event(event)
        for msg in channel_messages:
            if self.debug:
                self.logger.info(msg)
            else:
                self.logger.info("Minette> {}".format(
                    msg.text if hasattr(msg, "text") else msg.alt_text
                    if hasattr(msg, "alt_text") else msg.type))
        self.api.reply_message(token, channel_messages)

    def _extract_token(self, event):
        """
        Extract token from event

        Parameters
        ----------
        event : Event
            Event from LINE Messaging API

        Returns
        -------
        message : Message
            Request message object
        """
        return event.reply_token if hasattr(event, "reply_token") else ""

    def _to_minette_message(self, event):
        """
        Convert LINE Event object to Minette Message object

        Parameters
        ----------
        event : Event
            Event from LINE Messaging API

        Returns
        -------
        message : Message
            Request message object
        """
        msg = Message(
            type=event.type,
            token=event.reply_token if hasattr(event, "reply_token") else None,
            channel="LINE",
            channel_detail="Messaging",
            channel_user_id=event.source.user_id,
            channel_message=event,
            timestamp=datetime.now(self.timezone))
        if event.source.type in ["group", "room"]:
            if event.source.type == "group":
                msg.group = Group(id=event.source.group_id, type="group")
            elif event.source.type == "room":
                msg.group = Group(id=event.source.room_id, type="room")
        if isinstance(event, MessageEvent):
            msg.id = event.message.id
            msg.type = event.message.type
            if isinstance(event.message, TextMessage):
                msg.text = event.message.text
            elif isinstance(event.message, (
                    ImageMessage, VideoMessage, AudioMessage)):
                msg.payloads.append(
                    Payload(
                        content_type=event.message.type,
                        url="https://api.line.me/v2/bot/message/%s/content" %
                            event.message.id,
                        headers={
                            "Authorization": "Bearer {%s}" %
                            self.channel_access_token
                        })
                )
            elif isinstance(event.message, LocationMessage):
                msg.payloads.append(
                    Payload(content_type=event.message.type, content={
                        "title": event.message.title,
                        "address": event.message.address,
                        "latitude": event.message.latitude,
                        "longitude": event.message.longitude
                    })
                )
            elif isinstance(event.message, StickerMessage):
                msg.payloads.append(
                    Payload(content_type=event.message.type, content={
                        "package_id": event.message.package_id,
                        "sticker_id": event.message.sticker_id,
                    })
                )
        elif isinstance(event, PostbackEvent):
            msg.payloads.append(
                Payload(content_type="postback", content={
                    "data": event.postback.data,
                    "params": event.postback.params
                })
            )
        elif isinstance(event, FollowEvent):
            pass
        elif isinstance(event, UnfollowEvent):
            pass
        elif isinstance(event, JoinEvent):
            pass
        elif isinstance(event, LeaveEvent):
            pass
        elif isinstance(event, MemberJoinedEvent):
            pass
        elif isinstance(event, MemberLeftEvent):
            pass
        elif isinstance(event, BeaconEvent):
            pass
        else:
            pass
        return msg

    @staticmethod
    def _to_channel_message(message):
        """
        Convert Minette Message object to LINE SendMessage object

        Parameters
        ----------
        response : Message
            Response message object

        Returns
        -------
        response : SendMessage
            SendMessage object for LINE Messaging API
        """
        payload = next(iter([p for p in message.payloads
            if p.content_type != "quick_reply"]), None)
        quick_reply = next(iter([p.content for p in message.payloads
            if p.content_type == "quick_reply"]), None)
        if message.type == "text":
            return TextSendMessage(text=message.text,
                                   quick_reply=quick_reply)
        elif message.type == "image":
            return ImageSendMessage(original_content_url=payload.url,
                                    preview_image_url=payload.thumb,
                                    quick_reply=quick_reply)
        elif message.type == "audio":
            return AudioSendMessage(original_content_url=payload.url,
                                    duration=payload.content["duration"],
                                    quick_reply=quick_reply)
        elif message.type == "video":
            return VideoSendMessage(original_content_url=payload.url,
                                    preview_image_url=payload.thumb,
                                    quick_reply=quick_reply)
        elif message.type == "location":
            cont = payload.content
            return LocationSendMessage(title=cont["title"],
                                       address=cont["address"],
                                       latitude=cont["latitude"],
                                       longitude=cont["longitude"],
                                       quick_reply=quick_reply)
        elif message.type == "sticker":
            return StickerSendMessage(package_id=payload.content["package_id"],
                                      sticker_id=payload.content["sticker_id"],
                                      quick_reply=quick_reply)
        elif message.type == "imagemap":
            return ImagemapSendMessage(alt_text=message.text,
                                       base_url=payload.url,
                                       base_size=payload.content["base_size"],
                                       actions=payload.content["actions"],
                                       quick_reply=quick_reply)
        elif message.type == "template":
            return TemplateSendMessage(alt_text=message.text,
                                       template=payload.content,
                                       quick_reply=quick_reply)
        elif message.type == "flex":
            return FlexSendMessage(alt_text=message.text,
                                   contents=payload.content,
                                   quick_reply=quick_reply)
        else:
            return None



""" Base class for UserStore """
from abc import ABC, abstractmethod
import traceback
from logging import getLogger
from datetime import datetime
from pytz import timezone as tz

import objson

from ..models import User


class UserStore(ABC):
    """
    Base class for UserStore to read/write user information

    Attributes
    ----------
    config : minette.Config
        Configuration
    timezone : pytz.timezone
        Timezone
    logger : logging.Logger
        Logger
    table_name : str
        Database table name for read/write user data
    sqls : dict
        SQLs used in ContextStore
    """

    def __init__(self, config=None, timezone=None, logger=None,
                 table_name="user", **kwargs):
        """
        Parameters
        ----------
        config : minette.Config, default None
            Configuration
        timezone : pytz.timezone, default None
            Timezone
        logger : logging.Logger, default None
            Logger
        table_name : str, default "user"
            Database table name for read/write user data
        """
        self.config = config
        self.timezone = timezone or (
            tz(config.get("timezone", default="UTC")) if config else tz("UTC"))
        self.logger = logger if logger else getLogger(__name__)
        self.table_name = table_name
        self.sqls = self.get_sqls()

    @abstractmethod
    def get_sqls(self):
        pass

    def prepare_table(self, connection, prepare_params=None):
        """
        Check and create table if not exist

        Parameters
        ----------
        connection : Connection
            Connection for prepare

        query_params : tuple, default tuple()
            Query parameters for checking table
        """
        cursor = connection.cursor()
        cursor.execute(self.sqls["prepare_check"], prepare_params or tuple())
        if not cursor.fetchone():
            cursor.execute(self.sqls["prepare_create"])
            connection.commit()
            return True
        else:
            return False

    def get(self, channel, channel_user_id, connection):
        """
        Get user by channel and channel_user_id

        Parameters
        ----------
        channel : str
            Channel
        channel_user_id : str
            Channel user ID
        connection : Connection
            Connection

        Returns
        -------
        user : minette.User
            User
        """
        user = User(channel=channel, channel_user_id=channel_user_id)
        if not channel_user_id:
            return user
        try:
            cursor = connection.cursor()
            cursor.execute(self.sqls["get_user"], (channel, channel_user_id))
            row = cursor.fetchone()
            if row is not None:
                # convert to dict
                if isinstance(row, dict):
                    record = row
                else:
                    record = dict(
                        zip([column[0] for column in cursor.description], row))
                # convert type
                record["data"] = objson.loads(record["data"])
                # restore user
                user.id = record["user_id"]
                user.name = record["name"]
                user.nickname = record["nickname"]
                user.profile_image_url = record["profile_image_url"]
                user.data = record["data"] if record["data"] else {}
            else:
                cursor.execute(self.sqls["add_user"], (
                    channel, channel_user_id, user.id,
                    datetime.now(self.timezone), user.name, user.nickname,
                    user.profile_image_url, None)
                )
                connection.commit()
        except Exception as ex:
            self.logger.error(
                "Error occured in restoring user from database: "
                + str(ex) + "\n" + traceback.format_exc())
        return user

    def save(self, user, connection):
        """
        Save user

        Parameters
        ----------
        user : minette.User
            User to save
        connection : Connection
            Connection
        """
        user_dict = user.to_dict()
        serialized_data = objson.dumps(user_dict["data"])
        cursor = connection.cursor()
        cursor.execute(self.sqls["save_user"], (
            datetime.now(self.timezone), user.name, user.nickname,
            user.profile_image_url, serialized_data, user.channel,
            user.channel_user_id)
        )
        connection.commit()



from uuid import uuid4
from objson import Serializable


class User(Serializable):
    """
    User

    Attributes
    ----------
    id : str
        User ID
    name : str
        User name
    nickname : str
        Nickname
    channel : str
        Channel
    channel_user_id : str
        Channel user ID
    data : dict
        User data
    """
    def __init__(self, channel=None, channel_user_id=None):
        """
        Parameters
        ----------
        channel : str, default None
            Channel
        channel_user_id : str, default None
            Channel user ID
        """
        self.id = str(uuid4())
        self.name = ""
        self.nickname = ""
        self.channel = channel
        self.channel_user_id = \
            channel_user_id if isinstance(channel_user_id, str) else ""
        self.profile_image_url = ""
        self.data = {}
